﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using System.Collections.Generic;
using UnityEngine;

namespace PuppetMaster
{
    public enum AttackState
    {
        ready,
        windup,
        hit,
        final,
    }


    public class Attack
    {
        public Puppet Puppet = null;
        public AttackState state      = AttackState.ready;
        public AttackIds CurrentAttack;
        public MoveTypes CurrentMove;
        public Thing thing;

        public PersonBehaviour Enemy;
        public Vector2 EnemyTarget;
        public Dictionary<string, Rigidbody2D> XRB2 = new Dictionary<string, Rigidbody2D>();

        public float floatemp1;
        public float floatemp2;
        public float power      = 0f;
        public float facing     = 1f;
        public int frame        = 0;

        public MoveSet combatPose = new MoveSet("UpperArmFront:-11.68893:10,LowerArmFront:-96.55749:10");
        public List<MoveSet> AttackMoves = new List<MoveSet>();

        public enum AttackIds
        {
            club,
            thrust,
            thrustx,
        }

        public enum MoveTypes
        {
            stab,
            slash,
        }

        public Vector2 Vectemp1;
        public Vector2 Vectemp2;




        //
        // ─── TARGET ENEMY ────────────────────────────────────────────────────────────────
        //
        public void TargetEnemy()
        {
            PersonBehaviour TheChosen = (PersonBehaviour)null;
            PersonBehaviour[] people  = GameObject.FindObjectsOfType<PersonBehaviour>();
            floatemp1                 = float.MaxValue;

            bool lastKnockedOut       = false;

            foreach (PersonBehaviour person in people)
            {
                if (!person.isActiveAndEnabled || !person.IsAlive() || person == Puppet.PBO) continue;

                Vectemp1 = person.transform.GetComponentInChildren<Rigidbody2D>().position;

                if (Puppet.FacingLeft  && Vectemp1.x > Puppet.RB2["Head"].position.x) continue;
                if (!Puppet.FacingLeft && Vectemp1.x < Puppet.RB2["Head"].position.x) continue;

                floatemp2 = (thing.R.position - Vectemp1).sqrMagnitude;

                if (floatemp2 < floatemp1)
                {
                    floatemp1 = floatemp2;
                    TheChosen = person;
                    if (person.Consciousness < 1 || person.ShockLevel > 0.3f)
                    {
                        lastKnockedOut = true;
                    }
                } else if(lastKnockedOut)
                {
                    if (person.Consciousness >= 1 && person.ShockLevel < 0.3f)
                    {
                        if (floatemp2 - floatemp1 < 1f)
                        {
                            lastKnockedOut = false;
                            floatemp1 = floatemp2;
                            TheChosen = person;
                        }
                    }
                }
            }

            if (TheChosen == null) Enemy = TheChosen;

            //if (TheChosen == Enemy) return;

            Enemy = TheChosen;

            string[] ValidTargets = new string[]
            {
                "Head",
                "UpperBody",
                "MiddleBody",
                "LowerBody",
                "UpperArm",
                "UpperArmFront",
                "LowerArm",
                "LowerArmFront",
            };

            string randomTarget = ValidTargets[UnityEngine.Random.Range(0,ValidTargets.Length - 1)];  
            

            XRB2.Clear();

            Rigidbody2D[] RBs = Enemy.transform.GetComponentsInChildren<Rigidbody2D>();


            foreach (Rigidbody2D rb in RBs)
            {
                XRB2.Add(rb.name, rb);
            }

            EnemyTarget = XRB2[ randomTarget ].position;

        }

        //
        // ─── ATTACK INIT ────────────────────────────────────────────────────────────────
        //
        public void Init(AttackIds attackId)
        {

            if (state != AttackState.ready) return;

            thing                   = Puppet.HoldingF;
            thing.P.ForceContinuous = true;
            thing.AttackDamage      = 0f;
            frame                   = 0;
            power                   = 0;
            state                   = AttackState.windup;
            facing                  = Puppet.FacingLeft ? -1 : 1;
            state                   = AttackState.windup;

            thing.AttackedList.Clear();

            AttackMoves.Clear();

            CurrentAttack = attackId;

            if (attackId == AttackIds.club)
            {
                TargetEnemy();
                AttackMoves.Add(new MoveSet(@"
                        UpperArmFront:169.3435:10,
                        LowerArmFront:-43.21984:10"
                ));

                AttackMoves.Add(new MoveSet(@"
                        UpperBody:0.6148212:10,
                        UpperArmFront:286.4763:10,
                        LowerArmFront:0.4966094:10
                    "));
            }

            if (attackId == AttackIds.thrust)
            {
                TargetEnemy();
                AttackMoves.Add(new MoveSet(@"
                        UpperArmFront:112.2979:20,
                        UpperArm:-0.0164523:20,
                        LowerArm:-0.04736875:20,
                        LowerArmFront:-69.02728:20
                    "));
                AttackMoves.Add(new MoveSet(@"
                        UpperBody:-2.225564:20,
                        UpperArmFront:-12.23704:20,
                        UpperArm:-0.5381694:20,
                        LowerBody:2.078097:20,
                        LowerArm:-4.081913:20,
                        LowerArmFront:-6.531116:20
                    "));
            }
        }


        //
        // ─── ATTACK GO ────────────────────────────────────────────────────────────────
        //
        public void Go()
        {
            frame++;
            if (KB.ActionHeld || KB.Action2Held) power++;

            if (CurrentAttack == AttackIds.club) MeleeClub();
            if (CurrentAttack == AttackIds.thrust) MeleeThrust();
        }


        // - - - - - - - - - - - - - - - - - -
        //  MELEE:  CLUB
        // - - - - - - - - - - - - - - - - - -
        public void MeleeClub()
        {
            if (state == AttackState.windup)
            {
                if (Enemy != null)
                {
                    float distance = 0.1f;

                    if (XRB2.ContainsKey("UpperBody"))
                    {
                        distance = Mathf.Abs(XRB2["UpperBody"].position.x - Puppet.RB2["UpperBody"].position.x);

                    }

                    if (distance > 3.5f)
                    {
                        Enemy = (PersonBehaviour)null;
                    }
                    else if (!Puppet.IsCrouching && distance < 4.5f && distance > 1.0f)
                    {
                        Puppet.PBO.OverridePoseIndex = (int)PoseState.Walking;

                        Puppet.RB2["LowerLeg"].AddForce(Vector2.up + (Vector2.right * facing) * Puppet.TotalWeight * distance * 2f);
                        Puppet.RB2["UpperBody"].AddForce(Vector2.up + (Vector2.right * facing) * Puppet.TotalWeight * distance * 2f);
                        Puppet.RB2["Head"].AddForce(Vector2.up * Puppet.TotalWeight * distance * 2f);
                    }

                    if (!Puppet.IsCrouching && distance < 1.9f)
                    {
                        Puppet.PBO.OverridePoseIndex = 0;
                        Puppet.RB2["UpperBody"].velocity *= 0.001f;
                    }
                }
                else
                {
                    Puppet.RB2["Head"].AddForce(Vector2.right * facing * Puppet.TotalWeight * 2f);
                    Puppet.RB2["FootFront"].AddForce(Vector2.right * -facing * Puppet.TotalWeight * 1f);
                    Puppet.RB2["Foot"].AddForce(Vector2.right * -facing * Puppet.TotalWeight * 1f);

                    EnemyTarget = Puppet.RB2["UpperBody"].position + Vector2.right * facing;
                    
                }

                if (frame == 1)
                {
                    AttackMoves[0].neverEnding = Time.time + 1.0f;
                    CurrentMove= MoveTypes.slash;
                }


                if (AttackMoves[0].RunMove(Puppet))
                {
                    state = AttackState.hit;
                    frame = 0;
                    return;
                }

                if (!KB.ActionHeld && !KB.Action2Held)
                {
                    state = AttackState.hit;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.hit)
            {

                if (frame == 1)
                {
                    thing.AttackDamage = power;

                    //  Set attack damage power curve
                    if (power > 40) thing.AttackDamage = 30 - (power % 40);

                    

                    AttackMoves[1].neverEnding = Time.time + 2.0f;
                    Puppet.RunRigids(BodyInertia);
                }

                if (frame == 2)
                {
                    thing.R.AddForce((Vector2.right * facing) + (Vector2.down * 0.5f) * thing.AttackDamage * 2);
                    thing.R.AddTorque(thing.AttackDamage * -facing);

                    Puppet.RB2["LowerArmFront"].AddForce(Vector2.right * facing * Puppet.TotalWeight * 10f);
                    Puppet.RB2["UpperBody"].AddForce(Vector2.right * -facing * Puppet.TotalWeight * 15f);
                    Puppet.RB2["UpperArmFront"].AddTorque(20f * -facing * Puppet.TotalWeight);
                }

                if (frame > 2)
                {
                    float distance = thing.R.position.x - EnemyTarget.x;
                    //float distance = thing.R.position.x - XRB2["Head"].position.x;

                    if (Mathf.Abs(distance) < 2f)
                    {
                        thing.R.mass    = 0.1f;
                        //Vector2 target1 = (XRB2["Head"].position - thing.R.position);
                        Vector2 target1 = (EnemyTarget - thing.R.position);
                        float force     = 0.01f;// thing.canSlice ? 1.5f : 0.5f;
                        thing.R.AddForceAtPosition(target1 * force, (Vector2)((Vector2)thing.tr.localPosition + (Vector2)(thing.R.transform.right * facing * thing.size)), ForceMode2D.Impulse);

                        //Vectemp1 = thing.R.velocity;
                        //Vectemp1.y *= 0.05f;
                        //
                        //thing.R.velocity = Vectemp1;

                        Puppet.RB2["LowerArmFront"].AddForce(target1 * thing.AttackDamage * Puppet.TotalWeight);

                        Puppet.RB2["UpperBody"].velocity *= 0.001f;
                    }
                }


                //if (frame > 2 && frame < 8) thing.R.velocity *= 1.5f;

                if (frame == 4) thing.R.mass = 0.05f;

                //if (frame > 3 && !KB.ActionHeld && !KB.Action2Held)
                //{
                //    state = AttackState.final;
                //    frame = 0;
                //}

                //if (AttackMoves[1].RunMove(Puppet))
                if (frame > 15)
                {
                    state = AttackState.final;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.final)
            {
                thing.AttackDamage = 0f;
                thing.P.MakeWeightless();

                state = AttackState.ready;
                Puppet.RunRigids(BodyInertiaFix);
                Puppet.PBO.OverridePoseIndex = KB.Down ? Puppet.PuppetPose["JT_Duck"] : 0;
            }
        }

        // - - - - - - - - - - - - - - - - - -
        //  MELEE:  THRUST
        // - - - - - - - - - - - - - - - - - -
        public void MeleeThrust()
        {
            if (state == AttackState.windup)
            {
                if (frame == 1)
                {
                    AttackMoves[0].neverEnding = Time.time + 1.0f;
                    CurrentMove = MoveTypes.stab;
                }

                if (Enemy != null)
                {
                    float distance = 0.1f;

                    if (XRB2.ContainsKey("UpperBody"))
                    {
                        distance = Mathf.Abs(XRB2["UpperBody"].position.x - Puppet.RB2["UpperBody"].position.x);

                    }
                    if (distance > 3.5f)
                    {
                        Enemy = (PersonBehaviour)null;
                    }
                    else if (distance < 3.5f && distance > 1.4f)
                    {
                        Puppet.PBO.OverridePoseIndex = (int)PoseState.Walking;
                        Puppet.RB2["LowerLeg"].AddForce(Vector2.up + (Vector2.right * facing) * Puppet.TotalWeight * distance * 2f);
                        Puppet.RB2["UpperBody"].AddForce(Vector2.up + (Vector2.right * facing) * Puppet.TotalWeight * distance * 2f);
                        Puppet.RB2["Head"].AddForce(Vector2.up * Puppet.TotalWeight * distance * 2f);
                    }
                    if (distance < 1.4f)
                    {
                        Puppet.PBO.OverridePoseIndex = 0;
                        Puppet.RB2["UpperBody"].velocity *= 0.001f;
                    }
                } else
                {
                    EnemyTarget = Puppet.RB2["UpperBody"].position + Vector2.right * facing;
                    
                }

                if (AttackMoves[0].RunMove(Puppet))
                {
                    state = AttackState.hit;
                    frame = 0;
                }

                if (!KB.ActionHeld && !KB.Action2Held)
                {
                    state = AttackState.hit;
                    frame = 0;
                }


                return;
            }

            if (state == AttackState.hit)
            {
                

                if (frame == 1)
                {

                    thing.AttackDamage = power;

                    //  Set attack damage power curve
                    if (power > 40) thing.AttackDamage = 30 - (power % 40);

                    
                    AttackMoves[1].neverEnding = Time.time + 1.0f;
                }

                if (frame >= 5 && frame <= 10)
                {
                    if (Enemy != null)
                    {
                        //float distance = thing.R.position.x - XRB2["UpperBody"].position.x;
                        float distance = thing.R.position.x - EnemyTarget.x;

                        if (Mathf.Abs(distance) < 2f)
                        {
                            thing.R.mass    = 0.1f;
                            Vector2 target1 = (EnemyTarget - thing.R.position);
                            //Vector2 target1 = (XRB2["LowerBody"].position - thing.R.position);

                            thing.R.AddForceAtPosition(target1 * 1.5f, (Vector2)((Vector2)thing.tr.localPosition + (Vector2)(thing.R.transform.right * facing * thing.size)), ForceMode2D.Impulse);

                            Vectemp1 = thing.R.velocity;
                            Vectemp1.y *= 0.05f;

                            thing.R.velocity = Vectemp1;

                            Puppet.RB2["LowerArmFront"].AddForce(target1 * 50f * Puppet.TotalWeight);

                            Puppet.RB2["UpperBody"].velocity *= 0.001f;
                        }
                    }
                }

                if (Enemy == null)
                {
                    if (AttackMoves[1].RunMove(Puppet))
                    {
                        state = AttackState.final;
                        frame = 0;
                    }
                }
                else if (frame > 30)
                {
                    state = AttackState.final;
                    frame = 0;
                }

                if (frame > 6 && !KB.ActionHeld && !KB.Action2Held)
                {
                    state = AttackState.final;
                    frame = 0;
                }

                return;
            }

            if (state == AttackState.final)
            {
                thing.AttackDamage = 0f;
                thing.P.MakeWeightless();

                state = AttackState.ready;
                Puppet.RunRigids(BodyInertiaFix);

                Puppet.PBO.OverridePoseIndex = KB.Down ? Puppet.PuppetPose["JT_Duck"] : 0;
            }
        }

        public void BodyInertia(Rigidbody2D rb)
        {
            string exclude = "LowerArmFront UpperArmFront";
            if (exclude.Contains(rb.name)) return;
            rb.inertia = 0.5f;
        }

        public void BodyInertiaFix(Rigidbody2D rigid)
        {
            Puppet.SetRigidOriginal(rigid.name);
        }

        //
        // ─── MOVESET CLASS ────────────────────────────────────────────────────────────────
        //
        public class MoveSet
        {
            public float neverEnding                  = 0;
            public float lastOffset                   = 0f;
            public int framesOK                       = 0;
            private readonly List<LimbMove> moves     = new List<LimbMove>();
            public struct LimbMove
            {
                public string limbName;
                public float angle;
                public float force;
            }

            public MoveSet(string movesList)
            {
                Import(movesList);
            }

            public void Import(string movesList)
            {
                string[] moveItem = movesList.Trim().Split(',');
                foreach (string item in moveItem)
                {
                    if (item.Trim() == "") continue;

                    string[] parts = item.Trim().Split(':');
                    LimbMove move  = new LimbMove()
                    {
                        limbName = parts[0]
                    };
                    if (float.TryParse(parts[1], out float angle)) move.angle = angle;
                    if (float.TryParse(parts[2], out float force)) move.force = force;

                    moves.Add(move);
                }
            }

            public bool RunMove(Puppet Puppet)
            {
                float finishedPct = 0f;

                foreach (LimbMove move in moves)
                {
                    Puppet.LB[move.limbName].InfluenceMotorSpeed(Mathf.DeltaAngle(
                        Puppet.LB[move.limbName].Joint.jointAngle, move.angle * Puppet.PBO.AngleOffset) * move.force);

                    finishedPct += Mathf.Abs(Puppet.LB[move.limbName].Joint.jointAngle - move.angle);
                }

                float currentOffset = finishedPct / moves.Count;

                if (Mathf.Abs(currentOffset - lastOffset) < 0.1f)
                {
                    if (++framesOK >= 3)
                    {
                        lastOffset = 0f;
                        framesOK   = 0;

                        return true;
                    }
                }
                lastOffset = currentOffset;
                return (bool)(neverEnding > 0 && Time.time > neverEnding);

            }
        }


    }

}
