﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using UnityEngine;

namespace PuppetMaster
{
    public enum DiveState
    {
        ready,
        start,
        diving,
        landed,
    };

    public class Dive
    {
        public DiveState state        = DiveState.ready;
        public Puppet Puppet = null;
        public bool DoMaxPayne        = false;
        public int frame              = 0;
        public float force            = 0;
        public float facing           = 1;


        //
        // ─── DIVE INIT ────────────────────────────────────────────────────────────────
        //
        public void Init()
        {

            if (state == DiveState.ready)
            {
                frame = 0;

                Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Dive_1"];

                state = DiveState.start;
                force = 0;
                frame = 0;

                facing = Puppet.FacingLeft ? 1 : -1;
                if (KB.Modifier)
                {
                    Util.MaxPayne(true);
                    DoMaxPayne = true;
                }

                Puppet.Invincible(true);
                Puppet.DisableMoves = Time.time + 2.5f;

            }
        }

        //
        // ─── DIVE GO ────────────────────────────────────────────────────────────────
        //
        public void Go()
        {
            frame++;

            Vector2 direction = Puppet.RB2["UpperBody"].position;

            direction.y = 5.1f;
            direction.x = 12 * -facing;

            if (state == DiveState.start)
            {
                Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Dive_1"];

                if (Mathf.Abs(Puppet.LB["UpperLeg"].Joint.jointAngle) > 63f)
                {
                    Puppet.RunRigids(LightDrag);

                    Puppet.RB2["UpperBody"].AddForce(direction * (20 * Puppet.TotalWeight));
                    Puppet.RB2["LowerArm"].AddForce(Vector2.up * (20 * Puppet.TotalWeight));
                    Puppet.RB2["LowerArmFront"].AddForce(Vector2.up * (20 * Puppet.TotalWeight));
                    Puppet.RB2["UpperArm"].AddTorque(-facing * 40 * Puppet.TotalWeight);
                    Puppet.RB2["UpperArmFront"].AddTorque(-facing * 40 * Puppet.TotalWeight);

                    Puppet.PBO.OverridePoseIndex = 0;

                    state = DiveState.diving;
                    frame = 0;

                    return;
                }


                //if (frame < 10 ) Puppet.RB2["MiddleBody"].AddTorque(-facing * Puppet.TotalWeight * 1.1f);

                if (frame == 20)
                {
                    //Puppet.RB2["LowerBody"].AddForce(direction * (2 * Puppet.TotalWeight));
                }


            }


            if (state == DiveState.diving)
            {
                if (frame == 10) Puppet.PBO.OverridePoseIndex = 0;
                Puppet.RB2["LowerBody"].AddForce(direction * Puppet.TotalWeight);

                if (frame >= 5 && frame <= 7)
                {
                    Puppet.RB2["Head"].AddTorque(facing * Puppet.TotalWeight * 10);
                }
                if (frame > 14)
                {
                    frame = 0;
                    state = DiveState.landed;
                    //Puppet.RB2["LowerBody"].AddTorque(10f * facing * Puppet.TotalWeight);
                    Puppet.PBO.OverridePoseIndex = -1;
                }
            }




            if (state == DiveState.landed)
            {
                //if (frame > 5) Puppet.RunRigids(BodyInertia);

                if (Puppet.LB["Foot"].IsOnFloor || Puppet.LB["FootFront"].IsOnFloor)
                {

                    if (DoMaxPayne) Actions.AddTask(EndMaxPayne, 2.5f);

                    Puppet.RB2["MiddleBody"].AddTorque(50f * -facing * Puppet.TotalWeight);
                    Puppet.RB2["Head"].AddForce(Vector2.left * -facing * 20 * Puppet.TotalWeight);
                    Puppet.RB2["Foot"].AddForce(Vector2.down * 5 * Puppet.TotalWeight);
                    Puppet.RB2["FootFront"].AddForce(Vector2.down * 5 * Puppet.TotalWeight);
                    //Puppet.PBO.OverridePoseIndex = 0;

                    state = DiveState.ready;

                    Puppet.RunRigids(BodyInertia);

                    Actions.AddTask(finale, 1.5f);

                }
                if (Puppet.LB["LowerArm"].IsOnFloor || Puppet.LB["LowerArmFront"].IsOnFloor)
                {
                    Puppet.RB2["LowerBody"].AddTorque(10f * facing * Puppet.TotalWeight);
                }
            }
        }

        public void FixRigids(int seconds)
        {
            Actions.AddTask(ActionFixRigids, Time.time + seconds);
        }

        public void ActionFixRigids()
        {
            Puppet.Invincible(false);
            Puppet.RunRigids(BodyInertiaFix);
        }


        public void finale()
        {
            Puppet.Invincible(false);
            Puppet.RunLimbs(HealLimb);
            Puppet.RunRigids(BodyInertiaFix);
        }

        public void BodyInertiaFix(Rigidbody2D rigid)
        {
            Puppet.SetRigidOriginal(rigid.name);
        }

        public void LightDrag(Rigidbody2D rigid)
        {
            rigid.drag = 0.0001f;
        }
        public void BodyInertia(Rigidbody2D rigid)
        {
            rigid.drag = 0.1f;
        }
        public void HealLimb(LimbBehaviour limb)
        {
            limb.BruiseCount = 0;
        }
        public void EndMaxPayne()
        {
            Util.MaxPayne(false);
        }
    }

}
