﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using UnityEngine;

namespace PuppetMaster
{
    public enum JumpState
    {
        ready,
        start,
        launch,
        goingUp,
        goingDown,
        landed,
    };

    public class Jump
    {
        public int frame              = 0;
        public JumpState state        = JumpState.ready;
        public Puppet Puppet          = null;
        public float jumpStrength     = 0;
        public float torque           = 0;
        public float facing           = 1;

        public void MakeUninvincible() => Puppet.Invincible(false);


        public void Init()
        {

            if (state == JumpState.ready && !Puppet.JumpLocked && !Puppet.DisabledMoves)
            {
                Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_FrontFlip"];
                jumpStrength                 = 0;
                state                        = JumpState.start;
                frame                        = 0;
                torque                       = 0;
                facing                       = Puppet.FacingLeft ? -1 : 1;

                Puppet.BlockMoves = true;
            }
        }



        //
        // ─── JUMP GO ────────────────────────────────────────────────────────────────
        //
        public void Go()
        {

            frame++;

            Vector2 direction  = new Vector2(KB.Right ? 3 : KB.Left ? -3 : 1 * facing, 10);
            float keyDirection = KB.Left ? 1f : (KB.Right ? -1f : 0);

            if ((KB.Left && facing < 0) || (KB.Right && facing > 0)) direction.x *= 2;
            //else if  direction.x *= 2;

            if (state == JumpState.start)
            {
                if (KB.Up)
                {
                    jumpStrength += 0.01f;
                }
                else
                {
                    if (jumpStrength <= 0f)
                    {
                        state             = JumpState.ready;
                        Puppet.BlockMoves = false;

                        return;
                    }

                    Util.Notify("Jumping with Strength: " + jumpStrength, Util.VerboseLevels.Full);
                    if (jumpStrength > 3.5f) jumpStrength = 3.5f;

                    state = JumpState.launch;
                    frame = 0;
                    Puppet.Invincible(true);
                    return;

                }
            }

            if (state == JumpState.launch)
            {

                if ((!Puppet.LB["Foot"].IsOnFloor && !Puppet.LB["FootFront"].IsOnFloor)
                    || (Puppet.LB["Foot"].IsParalysed && Puppet.LB["FootFront"].IsParalysed))
                {
                    Puppet.PBO.OverridePoseIndex = -1;
                    Puppet.BlockMoves            = false;
                    state                        = JumpState.ready;

                    return;
                }


                Puppet.PBO.OverridePoseIndex = -1;

                if (KB.Left || KB.Right)
                {

                    torque = keyDirection * 10 * Puppet.TotalWeight * jumpStrength * 2f;
                    Puppet.RB2["MiddleBody"].AddTorque(torque, ForceMode2D.Force);
                    Puppet.RB2["UpperBody"].AddTorque(torque, ForceMode2D.Force);
                    Puppet.RB2["LowerBody"].AddTorque(torque, ForceMode2D.Force);
                }


                Puppet.RB2["Head"].AddRelativeForce(direction * (20 * (jumpStrength * Puppet.TotalWeight)));
                Puppet.RB2["UpperBody"].AddRelativeForce(direction * (20 * (jumpStrength * Puppet.TotalWeight)));
                Puppet.RB2["LowerBody"].AddRelativeForce(direction * (20 * (jumpStrength * Puppet.TotalWeight)));
                Puppet.RB2["Foot"].AddRelativeForce(direction * (20 * (jumpStrength * Puppet.TotalWeight)));
                Puppet.RB2["FootFront"].AddRelativeForce(direction * (20 * (jumpStrength * Puppet.TotalWeight)));

                if (KB.Left)
                {
                    Puppet.RB2["Head"].AddForce(Vector2.left * jumpStrength * Puppet.TotalWeight * 50);
                    Puppet.RB2["Foot"].AddForce(Vector2.right * jumpStrength * Puppet.TotalWeight * 50);
                    Puppet.RB2["FootFront"].AddForce(Vector2.right * jumpStrength * Puppet.TotalWeight * 50);
                }
                if (KB.Right)
                {
                    Puppet.RB2["Head"].AddForce(Vector2.right * jumpStrength * Puppet.TotalWeight * 50);
                    Puppet.RB2["Foot"].AddForce(Vector2.left * jumpStrength * Puppet.TotalWeight * 50);
                    Puppet.RB2["FootFront"].AddForce(Vector2.left * jumpStrength * Puppet.TotalWeight * 50);
                }
                state = JumpState.goingUp;
                frame = 0;
                return;
            }

            if (state == JumpState.goingUp)
            {
                if (frame < 10)
                {
                    Puppet.RB2["LowerBody"].AddForce(direction * (jumpStrength * Puppet.TotalWeight));
                }

                if (frame > 5 && Puppet.PBO.IsTouchingFloor)
                {
                    //Puppet.PBO.OverridePoseIndex = -1;

                    Puppet.BlockMoves = false;
                    state             = JumpState.ready;

                    if (jumpStrength < 0.2f) jumpStrength = 0.2f;
                    Puppet.DisableMoves = Time.time + jumpStrength / 2;

                    Actions.AddTask(MakeUninvincible, Time.time + 1);

                    return;
                }

                if (KB.Left || KB.Right)
                {
                    torque = keyDirection * 10 * Puppet.TotalWeight * jumpStrength * 0.32f;
                    Puppet.RB2["MiddleBody"].AddTorque(torque, ForceMode2D.Force);
                    Puppet.RB2["UpperBody"].AddTorque(torque, ForceMode2D.Force);
                    Puppet.RB2["LowerBody"].AddTorque(torque, ForceMode2D.Force);
                }
            }
        }
    }
}
