﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using UnityEngine;

namespace PuppetMaster
{
    public enum ProneState
    {
        ready,
        start,
        drop,
        final,
    };

    public class Prone
    {
        public ProneState state       = ProneState.ready;
        public Puppet Puppet = null;
        public float facing           = 1;
        public int frame              = 0;


        //
        // ─── PRONE INIT ────────────────────────────────────────────────────────────────
        //
        public void Init()
        {
            frame = 0;

            if (state == ProneState.ready && Puppet.PBO.OverridePoseIndex != Puppet.PuppetPose["JT_Prone"])
            {
                frame  = 0;
                state  = ProneState.start;
                facing = Puppet.FacingLeft ? -1 : 1;

                Puppet.RB2["UpperBody"].AddForce(Vector2.up * 5 * Puppet.TotalWeight);
            }
        }


        //
        // ─── PRONE GO ────────────────────────────────────────────────────────────────
        //
        public void Go()
        {
            frame++;

            if (state == ProneState.start)
            {
                Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Prone"];

                Puppet.RB2["UpperBody"].AddForce(Vector2.up * 25 * Puppet.TotalWeight);

                state = ProneState.drop;
                frame = 0;

                Puppet.IsCrouching = false;
                return;
            }

            if (state == ProneState.drop)
            {

                Puppet.RB2["UpperBody"].AddForce(Vector2.right * facing * 10 * Puppet.TotalWeight);
                Puppet.RB2["LowerLeg"].AddForce(Vector2.left * facing * 5 * Puppet.TotalWeight);
                Puppet.RB2["LowerLegFront"].AddForce(Vector2.left * facing * 5 * Puppet.TotalWeight);
                Puppet.RB2["MiddleBody"].AddTorque(-facing * 5 * Puppet.TotalWeight);
                Puppet.RB2["Foot"].AddForce(Vector2.up * 5 * Puppet.TotalWeight);

                if (frame >= 5)
                {
                    state = ProneState.final;
                    frame = 0;
                    return;
                }
            }

            if (state == ProneState.final)
            {
                if (Puppet.PBO.OverridePoseIndex != Puppet.PuppetPose["JT_Prone"])
                {
                    state = ProneState.ready;
                }
            }
        }
    }
}
