﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using System;
using UnityEngine;

namespace PuppetMaster
{
    public enum ThrowState
    {
        ready,
        aim,
        windup,
        fire,
    }

    public struct ThrownItem
    {
        public float EndRotation;
        public float StartRotation;
        public float Force;
        public float MaxTime;

        public Vector2 StartPos;
        public Vector2 EndPos;

        public Thing thing;
    }

    public class ThrowItem
    {
        public ThrowState state       = ThrowState.ready;
        public Puppet Puppet = null;
        public Thing heldThing        = null;
        
        public int frame              = 0;

        public float power            = 3.5f;
        public float facing           = 1f;
        private float initGravity;

        protected Material lineMaterial;

        private LineRenderer lr;

        private Vector2 ThrowStartPos;
        private Vector2 ThrowEndPos;

        public static bool isItemThrown = false;


        //
        // ─── THROW INIT ────────────────────────────────────────────────────────────────
        //
        public void Init()
        {
            frame = 0;

            if (state == ThrowState.ready)
            {
                if (Puppet.HoldingF == null) return;

                if (Puppet.IsAiming) Puppet.AimingStop();
                if (Puppet.Actions.combatMode) Puppet.Actions.combatMode = false;

                heldThing = Puppet.HoldingF;
                frame     = 0;
                state     = ThrowState.aim;
                facing    = Puppet.FacingLeft ? -1 : 1;

                Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Throw_x"];

                InitLine();

                Puppet.BlockMoves = true;

                initGravity = heldThing.P.InitialGravityScale;
                //initDrag    = heldThing.R.drag;
                
                if (Puppet.Actions.combatMode) Puppet.Actions.combatMode = false;
            }
        }


        //
        // ─── THROW GO ────────────────────────────────────────────────────────────────
        //
        public void Go()
        {
            if (heldThing == null)
            {
                Puppet.PBO.OverridePoseIndex = 0;
                lr.enabled                   = false;
                state                        = ThrowState.ready;
                Puppet.BlockMoves            = false;
                return;
            }

            frame++;

            if (state == ThrowState.aim)
            {
                if (KB.MouseDown || KB.Action) heldThing.P.SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);

                if (!KB.Throw)
                {
                    frame = 0;
                    state = ThrowState.windup;
                    return;
                }

                if (frame > 10)
                {
                    if (!lr.enabled)
                    {
                        lr.enabled    = true;
                        ThrowStartPos = Puppet.RB2["Head"].position;
                    }

                    lr.enabled = true;

                    Vector2 mousePo = (Vector2)Camera.main.ScreenToWorldPoint(Input.mousePosition);

                    if (facing == 1f  && mousePo.x < Puppet.RB2["Head"].position.x) { lr.enabled = false; return; }
                    if (facing == -1f && mousePo.x > Puppet.RB2["Head"].position.x) { lr.enabled = false; return; }


                    ThrowEndPos          = (Vector2)Camera.main.ScreenToWorldPoint(Input.mousePosition);
                    Vector2 _velocity    = (ThrowEndPos - ThrowStartPos) * power;
                    Vector2[] trajectory = Plot(ThrowStartPos, _velocity, 1000);
                    lr.positionCount     = trajectory.Length;
                    Vector3[] positions  = new Vector3[trajectory.Length];
                    int il               = positions.Length;

                    for (int i = 0; i < il; i++)
                    {
                        positions[i] = trajectory[i];
                    }

                    lr.SetPositions(positions);
                }

                return;
            }

            if (state == ThrowState.windup)
            {
                if (frame == 1)
                {
                    Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Throw_1"];
                }
                if (frame > 4)
                {
                    Puppet.PBO.OverridePoseIndex = -1;

                    Puppet.RB2["UpperArmFront"].AddTorque(5f * Puppet.TotalWeight * -facing);
                }
                if (frame > 9)
                {
                    Puppet.DropThing();
                }
                if (frame > 10)
                {
                    heldThing.R.velocity = Vector2.zero;

                    frame = 0;
                    state = ThrowState.fire;
                }
                return;
            }


            if (state == ThrowState.fire)
            {
                PuppetMaster.ChaseCam.QuickChase(heldThing, 3);

                Puppet.PBO.OverridePoseIndex    = 0;
                lr.enabled                      = false;
                state                           = ThrowState.ready;
                Puppet.BlockMoves               = false;

                Vector2 _velocity               = (ThrowEndPos - ThrowStartPos) * power;

                heldThing.R.velocity            = _velocity;
                
                if (heldThing.canStab || heldThing.isSyringe) { 
                    float currentRotation       = heldThing.R.rotation;
                    heldThing.R.angularVelocity = 0f;
                    heldThing.R.rotation        = currentRotation;

                    Actions.thrownItem          = new ThrownItem()
                    {
                        StartPos      = ThrowStartPos,
                        EndPos        = ThrowEndPos,
                        StartRotation = heldThing.R.rotation,
                        EndRotation   = Puppet.FacingLeft ? (90f * facing) + 360 + 180 : (95f * -facing) - 360,
                        Force         = _velocity.magnitude * power,
                        thing         = heldThing,
                        MaxTime       = Time.time + 1.5f,
                    };

                    Actions.thrownItem.EndRotation += heldThing.angleOffset * -facing;

                    isItemThrown      = true;
                }
            }
        }

        //
        // ─── THROW Plot ────────────────────────────────────────────────────────────────
        //
        public Vector2[] Plot(Vector2 pos, Vector2 velocity, int steps)
        {
            Vector2[] results    = new Vector2[steps];
            float timestep       = Time.fixedDeltaTime / Physics2D.velocityIterations;
            Vector2 gravityAccel = Physics2D.gravity * initGravity * timestep * timestep;
            //float drag           = 1f - timestep * initDrag;
            Vector2 moveStep     = velocity * timestep;
            RaycastHit2D hit;
            int i;
            for (i = 0; i < steps; i++)
            {
                moveStep += gravityAccel;

                pos += moveStep;
                if (i > 70) results[i - 71] = pos;
                if (i > 120)
                {
                    hit = Physics2D.Raycast(pos, (pos + moveStep).normalized, 0.5f);

                    if (hit.transform)
                    {

                        Array.Resize(ref results, i - 71);

                        return results;
                    }

                }



            }
            Array.Resize(ref results, i - 71);
            return results;
        }

        private void InitLine()
        {
            if (lr != null) return;

            lr = Puppet.PBO.gameObject.GetOrAddComponent<LineRenderer>();

            lr.enabled           = false;
            lr.startColor        = new Color(1f, 0.3f, 0f, 0.05f);
            lr.endColor          = new Color(1f, 0.4f, 0f, 0.05f);
            lr.startWidth        = 0.03f;
            lr.endWidth          = 0.01f;
            lr.numCornerVertices = 0;
            lr.numCapVertices    = 0;
            lr.useWorldSpace     = true;
            lr.alignment         = LineAlignment.View;
            lr.sortingOrder      = 2;
            lr.material          = ModAPI.FindMaterial("Sprites-Default");
            lr.textureMode       = LineTextureMode.RepeatPerSegment;
            lr.hideFlags         = HideFlags.HideAndDontSave;
        }

        public static void SpinThrownItem(ThrownItem thrown)
        {
            if (thrown.thing == null)
            {
                isItemThrown = false;
                return;
            }

            thrown.thing.R.SetRotation(Mathf.Lerp(thrown.thing.R.rotation, thrown.EndRotation, thrown.Force * Time.deltaTime * 0.11f));

            if (Mathf.Abs(thrown.thing.R.rotation - thrown.EndRotation) < 1) isItemThrown = false;
            else if (Time.time > thrown.MaxTime) isItemThrown = false;
        }
    }

}
