﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    
//                                           
using UnityEngine;

namespace PuppetMaster
{
    [SkipSerialisation]
    public class ChaseCam : MonoBehaviour
    {
        public static PhysicalBehaviour GPB;         
        public static PhysicalBehaviour ChaseTarget;
        public static PhysicalBehaviour LastTarget;  
        public static GameObject HiddenFocus;        

        public static int TargetHash     = 0;
        public static int LastHash       = 0;

        public float Yoffset             = 0;
        public float LastYoffset         = 0;
        public float maxMoveSpeed        = 7.1f;
        public float smoothTime          = 0.4f;
        public float minDistance         = 2f;
        public float qTime               = 0;
        public float qSeconds            = 0;
        public float sensitivity         = 7f;
        
        public float distance            = 0.0f;
        public float lastDelta           = 0.0f;
        public float deltaMag            = 0.0f;
        public float lockedtime          = 0f;

        public bool isChasing            = false;
        public bool isMoving             = false;
        public bool isLockedOn           = false;
        public bool qChaseActive         = false;
        public bool checkedMouse         = false;
        public bool targetClicked        = false;
        public bool qChaseGone           = false;

        public Vector2 lastPosition      = new Vector2();
        private Vector2 currentVelocity;


        public void Awake()
        {
            CreateChaseObject();
        }


        //
        // ─── UNITY LATE UPDATE ────────────────────────────────────────────────────────────────
        //
        public void LateUpdate()
        {
            if (!isChasing) return;

            if (KB.MouseDown)
            {
                if (!checkedMouse)
                {
                    checkedMouse = true;

                    Collider2D[] NoCollide = PuppetMaster.Puppet.PBO.transform.root.GetComponentsInChildren<Collider2D>();

                    foreach (Collider2D collider in NoCollide)
                    {
                        if (!collider) continue;

                        if (collider.OverlapPoint((Vector2)Global.main.MousePosition))
                        {
                            targetClicked = true;
                            break;
                        }
                    }
                }

                if (targetClicked) return;
            }

            else checkedMouse = targetClicked = false;

            DoChaseCam();

        }


        //
        // ─── UNITY UPDATE ────────────────────────────────────────────────────────────────
        //
        public void Update()
        {
            // ModAPI.Notify("LastDelta: " + lastDelta);
            if (qChaseActive)
            {
                if (Time.time > qTime)
                {
                    if (qSeconds == 0)
                    {
                        qChaseActive = false;
                        qTime        = 0;
                        Chase(LastTarget, LastHash, false);
                    }
                    else if (lastDelta < 0.05f)
                    {
                        qTime    = Time.time + qSeconds;
                        qSeconds = 0;
                    }
                }
                if (KB.AnyKey && !KB.Throw) StopQuickChase();

            }
        }


        //
        // ─── CHASE ────────────────────────────────────────────────────────────────
        //
        public void Chase(PhysicalBehaviour ctarget, int tHash, bool setYoffset=true)
        {
            if (!HiddenFocus || HiddenFocus.name != "ChaseCam") CreateChaseObject();

            ChaseTarget = ctarget;
            TargetHash  = tHash;

            Util.Notify("Chase Cam Targeting: " + ChaseTarget.name, Util.VerboseLevels.Full);

            HiddenFocus.transform.position = Global.CameraPosition;
            Yoffset = HiddenFocus.transform.position.y;
            Yoffset = setYoffset ? Global.main.CameraControlBehaviour.transform.position.y - ctarget.transform.position.y : LastYoffset;

            Global.main.CameraControlBehaviour.CurrentlyFollowing.Clear();
            Global.main.CameraControlBehaviour.CurrentlyFollowing.Add(GPB);

            isChasing = true;
        }


        //
        // ─── DO CHASE CAM ────────────────────────────────────────────────────────────────
        //
        public void DoChaseCam()
        {
            if (ChaseTarget == null || ChaseTarget.transform == null)
            {
                if (qChaseActive && !qChaseGone)
                {
                    qTime      = Time.time + 1.2f;
                    qSeconds   = 0f;
                    qChaseGone = true;

                    return;
                }

                else if (!qChaseActive) Stop();

                return;
            }

            Vector3 chasePosition = ChaseTarget.transform.position;

            float velo = ChaseTarget.rigidbody.velocity.magnitude * Time.deltaTime * 10;

            if (!qChaseActive) {
                chasePosition.x += 5f * -PuppetMaster.Puppet.Facing;
            }
            if (velo < 1f) chasePosition.y += Yoffset;



            distance     = (HiddenFocus.transform.position - chasePosition).sqrMagnitude;
            sensitivity  = isMoving ? 10.5f : 17f;
            maxMoveSpeed = Mathf.Max(5, ((lastDelta + velo) / 2) * distance);
            lastDelta    = velo;



            if (isMoving && velo < 0.01f && distance < sensitivity)
            {
                if (Time.time > lockedtime) lockedtime = Time.time + 2f;

                isMoving   = false;
                isLockedOn = false;
            }

            if (distance < sensitivity && Mathf.Abs(chasePosition.y - HiddenFocus.transform.localPosition.y) < 1f)
            {
                //isMoving = false;
                isLockedOn = false;
                return;
            }

            isMoving = true;

            // Offsets the target position so that the object keeps its distance.
            if (distance > 160f || isLockedOn)
            {
                isLockedOn = true;
                HiddenFocus.transform.position = Vector3.Slerp(HiddenFocus.transform.position, chasePosition, 25f * Time.deltaTime);
            }
            else
            {
                isLockedOn = false;
                HiddenFocus.transform.position = Vector2.SmoothDamp(HiddenFocus.transform.position, chasePosition, ref currentVelocity, Time.deltaTime, maxMoveSpeed);
            }
        }

        //
        // ─── CREATE CHASE OBJECT ────────────────────────────────────────────────────────────────
        //
        public void CreateChaseObject()
        {
            //  Create a hidden chase-cam object that the camera follows
            Texture2D tex   = new Texture2D(0, 0);
            Sprite mySprite = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.1f, 0.1f), 0.1f);
            HiddenFocus     = ModAPI.CreatePhysicalObject("ChaseCam", mySprite);

            // Initialize chase-cam object
            HiddenFocus.transform.position = Global.CameraPosition;

            GPB = HiddenFocus.GetOrAddComponent<PhysicalBehaviour>();

            GPB.MakeWeightless();
            GPB.Selectable = false;
            GPB.Deletable  = false;

            LayerSerialisationBehaviour LSB = HiddenFocus.GetOrAddComponent<LayerSerialisationBehaviour>();
            LSB.Layer = 10;

            PuppetMaster.ChaseCam = this;

        }


        //
        // ─── QUICK CHASE ────────────────────────────────────────────────────────────────
        //
        public void QuickChase(Thing thing, float seconds)
        {
            LastTarget  = ChaseTarget;
            LastHash    = TargetHash;
            LastYoffset = Yoffset;

            ChaseTarget = thing.P;
            TargetHash  = thing.Hash;

            HiddenFocus.transform.position = Global.CameraPosition;
            Yoffset = HiddenFocus.transform.position.y;

            Global.main.CameraControlBehaviour.CurrentlyFollowing.Clear();
            Global.main.CameraControlBehaviour.CurrentlyFollowing.Add(GPB);

            isChasing    = true;
            qChaseActive = true;
            qTime        = Time.time + 2f;
            qSeconds     = seconds;
            qChaseGone   = false;
        }

        public void StopQuickChase()
        {
            qChaseActive = false;
            qTime        = 0;
            Chase(LastTarget, LastHash, false);
        }


        //
        // ───  STOP CHASE ────────────────────────────────────────────────────────────────
        //
        public void StopChase(int tHash)
        {
            if (tHash == TargetHash)
            {
                ChaseTarget = null;
                TargetHash  = 0;
                isChasing   = false;
            }
        }

        public void Stop()
        {
            if (GPB == null)
            {
                this.enabled = false;
                return;
            }

            if (Global.main.CameraControlBehaviour.CurrentlyFollowing.Contains(GPB))
            {
                Global.main.CameraControlBehaviour.CurrentlyFollowing.Clear();
            }

            this.enabled = false;
        }

    }
}
