﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using System.Collections.Generic;
using UnityEngine;

namespace PuppetMaster
{
    [SkipSerialisation]
    public class Inventory
    {
        [SkipSerialisation] 
        public Puppet Puppet;
        [SkipSerialisation]
        public Thing Clone;

        public bool InventoryTriggered           = false;
        public bool InventoryDone                = false;
        public bool TriggerPickup                = false;

        public int CurrentSpot                    = -1;
        private int LastItemPos                   = -1;

        [SkipSerialisation] 
        private readonly Dictionary<int, Thing> MyInventory = new Dictionary<int, Thing>();

        public Inventory(Puppet puppet)
        {
            Puppet = puppet;
        }

        //
        // ─── INIT INVENTORY ────────────────────────────────────────────────────────────────
        //
        public void InitInventory()
        {
            KB.DisableNumberKeys(); 

            InventoryTriggered  = true;
            InventoryDone       = false;
            CurrentSpot         = -1;
        }

        public void DoInventory()
        {
            if (!InventoryTriggered) InitInventory();
            else
            {
                int spot = KB.CheckNumberKey();

                if (spot > -1 && CurrentSpot != spot) CurrentSpot = spot;

                if (KB.MouseDown)
                {
                    PhysicalBehaviour clickedPB = Util.GetClickedItem();

                    if (clickedPB != null && PuppetMaster.CanHoldItem(clickedPB))
                    {
                        Thing thing = clickedPB.gameObject.GetOrAddComponent<Thing>();
                        if (thing != null) StoreItem(thing);
                    }

                }
            }
        }


        //
        // ─── RUN INVENTORY ────────────────────────────────────────────────────────────────
        //
        public void RunInventory()
        {
            InventoryTriggered  = false;

            KB.EnableNumberKeys();

            if (InventoryDone) return;

            InventoryDone       = true;



            if ((Puppet.HoldingF == null || Puppet.HoldingF.Name == null) && CurrentSpot == -1)
            {
                if (LastItemPos > -1)
                {
                    SelectItem(LastItemPos);
                    return;
                }

                 Util.Notify("<color=orange>Hold </color>[ I ]<color=orange> and press </color>#[ 1-9 ]<color=orange> to store/retrieve Items</color>");
                 return;
            }

            if (Puppet.HoldingF != null) StoreItem(Puppet.HoldingF, -1);
            
            if (CurrentSpot > -1) SelectItem(CurrentSpot);
            else if (Puppet.HoldingF == null)
            {
                Puppet.Actions.combatMode = false;
                if (Puppet.IsAiming) Puppet.AimingStop();
            }
            
        }



        //
        // ─── STORE ITEM ────────────────────────────────────────────────────────────────
        //
        public bool StoreItem(Thing thing, int spot = -1)
        {

            if (Puppet.IsAiming) Puppet.AimingStop();
            
            //  Check if we already have blueprint of this item
            int curpos = LocateItem(thing);

            if (curpos > -1)
            {
                //  We already have the item blueprint, so destroy this one
                thing.Destroy();

                if (++curpos >= GetOpenInventorySpot())
                {
                    LastItemPos = 1;
                    return true;
                }

                LastItemPos = curpos;

                SelectItem(LastItemPos);

                return true;
            }


            if(spot == -1) spot = GetOpenInventorySpot();
            
            if (spot == -1)
            {
                Util.Notify("<color=orange>Inventory Full</color>");
                return false;
            }

            CleanItem(thing);
            
            MyInventory.Add(spot, thing);

            LastItemPos   = spot;
            InventoryDone = true;

            Util.Notify(thing.Name + " <color=orange>stored in position #:</color> " + spot, Util.VerboseLevels.Minimal);

            return true;

        }



        //
        // ─── GET OPEN INVENTORY SPOT ────────────────────────────────────────────────────────────────
        //
        public int GetOpenInventorySpot()
        {
            for (int i = 0; ++i <= 10;)
                if (!MyInventory.ContainsKey(i % 9)) return i;

            return -1;
        }


        //
        // ─── LOCATE ITEM ────────────────────────────────────────────────────────────────
        //
        public int LocateItem(Thing thing)
        {
            foreach (KeyValuePair<int, Thing> pair in MyInventory)
            {
                if (pair.Value.Hash == thing.Hash || pair.Value.Name == thing.Name) return pair.Key;
            }

            return -1;
        }


        //
        // ─── CLEAN ITEM ────────────────────────────────────────────────────────────────
        //
        public void CleanItem(Thing thing)
        {
            if (thing.PuppetGrip)
            {
                Puppet.DropThing();
            } 
            
            thing.BreakConnections();

            thing.G.layer              = 10;
            thing.G.transform.position = (new Vector2(-100, -100));
            thing.R.position           = (new Vector2(1, 1));
            thing.R.bodyType           = RigidbodyType2D.Static;

        }


        //
        // ─── SELECT ITEM ────────────────────────────────────────────────────────────────
        //
        public bool SelectItem(int spotChk)
        {
            Thing thing = (Thing)null;

            for (int i = spotChk; i <= 10; ++i)
            {
                if (MyInventory.TryGetValue(i % 9, out thing)) break;
            }
            
            if (thing == null) return false;

            GameObject instance = UnityEngine.Object.Instantiate(thing.G, thing.G.transform.position, Quaternion.identity);
            
            instance.name       = thing.Name;
            
            Clone               = instance.GetOrAddComponent<Thing>();
            Clone.R.bodyType    = RigidbodyType2D.Dynamic;
            Clone.G.layer       = 9;
            Clone.Name          = thing.Name;



            PuppetMaster.LastClickedItem = Clone.P;

            //Puppet.HoldThing(Clone);
            TriggerPickup       = true;
            InventoryDone       = true;
            LastItemPos         = spotChk;

            return true;

        }
       
    }

}