﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;

namespace PuppetMaster
{
    [SkipSerialisation]
    public class Puppet
    {
        [SkipSerialisation] public bool IsActive                 = false;
        [SkipSerialisation] public bool IsWalking                = false;
        [SkipSerialisation] public bool IsCrouching              = false;
        [SkipSerialisation] public bool IsAiming                 = false;
        [SkipSerialisation] public bool IsFiring                 = false;
        [SkipSerialisation] public bool JumpLocked               = false;
        [SkipSerialisation] public bool IsEmote                  = false;
        [SkipSerialisation] public bool FacingLeft               = false;
        [SkipSerialisation] public bool BlockMoves               = false;
        [SkipSerialisation] public bool checkFlipColliders       = false;
        [SkipSerialisation] public bool pauseAiming              = false;
        
        [SkipSerialisation] public int EmoteId                   = -1;
        
        [SkipSerialisation] public float Facing                  = -1f;
        [SkipSerialisation] public float TotalWeight             = 0;
        [SkipSerialisation] public float DisableMoves            = 0;
        
        [SkipSerialisation] public PersonBehaviour PBO           = null;
        [SkipSerialisation] public Inventory Inventory           = null;
        [SkipSerialisation] public Actions Actions               = null;

        [SkipSerialisation] public Thing HoldingF                = null;
        [SkipSerialisation] public Thing HoldingB                = null;

        [SkipSerialisation] public GripBehaviour GripF           = null;
        [SkipSerialisation] public GripBehaviour GripB           = null;

        [SkipSerialisation] public Dictionary<string, int> PuppetPose     = new Dictionary<string, int>();
        [SkipSerialisation] public Dictionary<string, Rigidbody2D> RB2    = new Dictionary<string, Rigidbody2D>();
        [SkipSerialisation] public Dictionary<string, LimbBehaviour> LB   = new Dictionary<string, LimbBehaviour>();
        private readonly Dictionary<string, LimbSnapshot> LimbOriginals   = new Dictionary<string, LimbSnapshot>();
        private readonly Dictionary<string, RigidSnapshot> RigidOriginals = new Dictionary<string, RigidSnapshot>();
        
        [SkipSerialisation] public List<Collider2D> FlipColliders = new List<Collider2D>();

        [SkipSerialisation]
        public struct LimbSnapshot
        {
            public float Health;
            public float Numbness;
            public float BreakingThreshold;
            public float Vitality;
            public float RegenerationSpeed;
            public float BaseStrength;
            public bool Broken;
            public bool Frozen;
            public ushort BruiseCount;
            public float jLimitMin;
            public float jLimitMax;
        }

        [SkipSerialisation]
        public class RigidSnapshot
        {
            [SkipSerialisation] public float inertia;
            [SkipSerialisation] public float mass;
            [SkipSerialisation] public float drag;
        }

        [SkipSerialisation] public enum Emotes {
            None,
            Waving,
            Victory,
            Prone,
            Dive1,
            Dive2,
        };



        [SkipSerialisation] public bool IsReady    => (bool)(PuppetMaster.Activated 
                                                   && PBO.isActiveAndEnabled 
                                                   && PBO.IsAlive() 
                                                   && PBO.Consciousness >= 1 
                                                   && PBO.ShockLevel < 0.3f );

        [SkipSerialisation] public bool IsUpright  => (bool)(!LB["LowerBody"].IsOnFloor
                                                   && !LB["UpperBody"].IsOnFloor
                                                   && !LB["Head"].IsOnFloor
                                                   && LB["FootFront"].IsOnFloor
                                                   && LB["Foot"].IsOnFloor);

        [SkipSerialisation] public bool CanWalk    => (bool)(PBO.IsTouchingFloor 
                                                   && !IsWalking 
                                                   && !IsCrouching
                                                   && !KB.Up 
                                                   && Actions.prone.state == ProneState.ready);
        [SkipSerialisation] public bool IsFlipped     => (bool)(checkFlip());
        [SkipSerialisation] public bool DisabledMoves => (bool)(BlockMoves || DisableMoves > Time.time);
        

        public bool checkFlip()
        {
            
            return (PBO.transform.localScale.x < 0.0f);
        }
        //
        // ─── UNITY AWAKE ────────────────────────────────────────────────────────────────
        //
        public void Init(PersonBehaviour _pbo)
        {
            TotalWeight       = 0f;
            PBO               = _pbo;

            //  Cache Rigidbody Maps
            Rigidbody2D[] RBs = PBO.transform.GetComponentsInChildren<Rigidbody2D>();
            
            foreach (Rigidbody2D rb in RBs)
            {

                //  Take snapshots of the current values for drag, intertia & mass
                //  Since these are modified during actions, we can set it back proper
                //
                RB2.Add(rb.name, rb);

                RigidSnapshot RBOG = new RigidSnapshot()
                {
                    drag    = rb.drag,
                    inertia = rb.inertia,
                    mass    = rb.mass,
                };

                RigidOriginals.Add(rb.name, RBOG);

                TotalWeight += rb.mass;

                GripBehaviour GB;

                //  @TODO: This shit is broken
                switch (rb.name)
                {
                    case "LowerArmFront":
                        if (rb.TryGetComponent<GripBehaviour>(out GB)) GripF = GB;
                        if (GripF.CurrentlyHolding != null) HoldingF = GripF.CurrentlyHolding.gameObject.GetOrAddComponent<Thing>();
                        break;

                    case "LowerArm":
                        if (rb.TryGetComponent<GripBehaviour>(out GB)) GripB = GB;
                        if (GripB.CurrentlyHolding != null) HoldingB = GripB.CurrentlyHolding.gameObject.GetOrAddComponent<Thing>();
                        break;
                }
            }

            //  Cache LimbBheaviour Maps
            LimbBehaviour[] LBs = PBO.GetComponentsInChildren<LimbBehaviour>();

            foreach (LimbBehaviour limb in LBs) 
            {
                LB.Add(limb.name, limb);

                LimbSnapshot LBOG = new LimbSnapshot()
                {
                    BaseStrength      = limb.BaseStrength,
                    BreakingThreshold = limb.BreakingThreshold,
                    Broken            = limb.Broken,
                    BruiseCount       = limb.BruiseCount,
                    Frozen            = limb.Frozen,
                    Health            = limb.Health,
                    Numbness          = limb.Numbness,
                    RegenerationSpeed = limb.RegenerationSpeed,
                    Vitality          = limb.Vitality,
                };

                if (limb.HasJoint)
                {
                    LBOG.jLimitMax = limb.Joint.limits.max;
                    LBOG.jLimitMin = limb.Joint.limits.min;
                }
                LimbOriginals.Add(limb.name, LBOG);
            }

            if (PBO.OverridePoseIndex != (int)PoseState.Rest || PBO.OverridePoseIndex != (int)PoseState.Sitting)
            {
                PBO.OverridePoseIndex = -1;
            }

            //  Assign custom poses
            PuppetPose  = CustomPoses.Assign(PBO);

            Actions        = new Actions(this);
            Inventory      = new Inventory(this);
            FacingLeft     = IsFlipped ? true : false;
            Facing         = FacingLeft ? 1f : -1f;


            RagdollPose pose = PBO.Poses[6];
            for (int i = pose.Angles.Count; --i >= 0;)
            {
                RagdollPose.LimbPose angle = pose.Angles[i];
                angle.RandomInfluence      = 0f;
                angle.RandomSpeed          = 0f;
                pose.Angles[i]         = angle;
            }

            PBO.Poses[6] = pose;
            PBO.Poses[6].DragInfluence = 1.2f;
            PBO.Poses[6].ConstructDictionary();

            Strengthen();

            Util.Notify("Puppet set: <color=yellow>" + PBO.name + "</color>");

            PhysicalBehaviour PBX = PBO.GetComponentInChildren<PhysicalBehaviour>();

            if (PBX != null) PuppetMaster.ChaseCam.Chase(PBX, this.GetHashCode());
        }
        //
        // ─── UNITY FIXED UPDATE ────────────────────────────────────────────────────────────────
        //
        public void FixedUpdate()
        {

            Actions.RunActions();


            if (IsAiming) { 
                if (!IsReady) IsAiming = false;
                else AimWeapon();
            }

            if (IsWalking)
            {
                if (KB.Left && !FacingLeft && Flip()) FacingLeft = true;
                else if (KB.Right && FacingLeft && Flip()) FacingLeft = false;
                else
                {
                    IsWalking = false;
                    StopPerson();
                }
            }

            if (IsCrouching)
            {

                if (!KB.Down)
                {
                    IsCrouching = false;
                    StopPerson();
                }

                else
                {
                    if (KB.Left || KB.Right)
                    {
                        if (FacingLeft != KB.Left && IsReady) Flip();
                    }
                }
            }

            if (Inventory.TriggerPickup)
            {
                Inventory.TriggerPickup = false;

                HoldThing(Inventory.Clone);
            }

            if (KB.Inventory) Inventory.DoInventory();

            if (Inventory.InventoryTriggered && !KB.Inventory) Inventory.RunInventory();

            if (checkFlipColliders && Time.frameCount % 5 == 0) CheckClearedCollisions();
        }


        //
        // ─── UNITY UPDATE ────────────────────────────────────────────────────────────────
        //
        public void Update()
        {
            if (KB.AnyKey && !IsReady) return;

            CheckControls();
        
        }



        //
        // ─── FLIP ────────────────────────────────────────────────────────────────
        //
        public bool Flip()
        {
            if (DisabledMoves || !IsUpright || (KB.Left && KB.Right) || Actions.jump.state != JumpState.ready ||
                !PBO.isActiveAndEnabled || !PBO.IsAlive() || !PBO.IsTouchingFloor || PBO.Consciousness < 1) return false;

            if (HoldingF != null)
            {
                foreach (PhysicalBehaviour.Penetration penetration in HoldingF.P.penetrations)
                {
                    if (penetration.Active) return false;
                }
            }

                //if (CheckFlipCollision()) return false;

           const string limbNamesList = "LowerBody,MiddleBody,UpperBody,UpperArm,UpperArmFront,Head";

            //  Take position snapshot of currently help items
            //ValidateHeldItems();

            //HoldingF?.SetHoldingPosition();
            //HoldingB?.SetHoldingPosition();

            Vector3 flipScale = PBO.transform.localScale;
            Vector3 moveA;
            Vector3 moveB;
            Vector2 moveDif;


            flipScale.x *= -1;

            moveB = RB2["Head"].transform.position;

            foreach (LimbBehaviour limb in LB.Values)
            {
                if (limb.HasJoint)
                {
                    limb.BreakingThreshold *= 8;

                    if (!limbNamesList.Contains(limb.name))
                    {
                        JointAngleLimits2D t = limb.Joint.limits;
                        t.min *= -1f;
                        t.max *= -1f;
                        limb.Joint.limits = t;
                        limb.OriginalJointLimits = new Vector2(limb.OriginalJointLimits.x * -1f, limb.OriginalJointLimits.y * -1f);
                    }
                }
            }

            PBO.transform.localScale = flipScale;

            moveA   = RB2["Head"].transform.position;
            moveDif = moveB - moveA;

            PBO.AngleOffset *= -1f;
            PBO.transform.position = new Vector2(PBO.transform.position.x + moveDif.x, PBO.transform.position.y);

            foreach (LimbBehaviour limb in PBO.Limbs) if (limb.HasJoint) limb.Broken = false;

            if (HoldingF != null)
            {

                LB["LowerArmFront"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                HoldingF.ResetPosition();

            }

            FacingLeft = IsFlipped;
            Facing     = FacingLeft ? 1f : -1f;

            HoldingB?.ResetPosition();

            CheckFlipCollision();

            return true;
        }

        public void CheckFlipCollision()
        {
            Collider2D[] noCollide = PBO.transform.root.GetComponentsInChildren<Collider2D>();
            bool isClear;
            Collider2D[] tmpCol;

            if (HoldingF != null)
            {
                Collider2D[] itemcollide = HoldingF.P.GetComponents<Collider2D>();

                foreach (Collider2D coll in itemcollide)
                {
                    coll.enabled = false;
                    isClear = true;

                    tmpCol = Physics2D.OverlapBoxAll(coll.transform.position, (Vector2)coll.bounds.size, coll.transform.eulerAngles.y);
                    if (tmpCol.Length > 0)
                    {
                        foreach (Collider2D col in tmpCol)
                        {
                            PersonBehaviour hitperson = col.attachedRigidbody.GetComponentInParent<PersonBehaviour>();
                            if (hitperson != null && hitperson != PBO)
                            {
                                checkFlipColliders = true;
                                FlipColliders.Add(col);
                                col.enabled = false;
                               // isClear = false;

                            }
                        }
                    }
                    coll.enabled = isClear;
                }

            }

            foreach (Collider2D coll in noCollide) {
                coll.enabled = false;
                isClear = true;

                tmpCol = Physics2D.OverlapBoxAll(coll.transform.position, (Vector2)coll.bounds.size, 0 );
                if (tmpCol.Length > 0) 
                {
                    foreach (Collider2D col in tmpCol) 
                    { 
                        PersonBehaviour hitperson = col.attachedRigidbody.GetComponentInParent<PersonBehaviour>();
                        if (hitperson != null && hitperson != PBO)
                        {
                            checkFlipColliders = true;
                            FlipColliders.Add(coll);
                            isClear = false;
                        }
                    }
                }
                coll.enabled = isClear;
            }

            return;
        }

        public bool CheckClearedCollisions()
        {
            Collider2D tmpCol;

            for (int i = FlipColliders.Count; --i >= 0;)
            {
                Collider2D coll = (Collider2D)FlipColliders[i];

                if (coll == null)
                {
                    FlipColliders.RemoveAt(i);
                    continue;
                }
                if (tmpCol = Physics2D.OverlapBox(coll.transform.position, (Vector2)coll.bounds.size, 0))
                {
                    PersonBehaviour hitperson = tmpCol.attachedRigidbody.GetComponentInParent<PersonBehaviour>();
                    if (hitperson == null || hitperson == PBO)
                    {
                        coll.enabled = true;
                        FlipColliders.RemoveAt(i);
                    }
                }
            }

            checkFlipColliders = FlipColliders.Count > 0;
            return checkFlipColliders;
        }

        //
        // ─── CHECK CONNTROLS ────────────────────────────────────────────────────────────────
        //
        public void CheckControls()
        {
            if (!KB.ActionHeld && !KB.MouseDown) IsFiring = false;
            if(KB.Action2 || KB.Action)
            {
                if (HoldingF != null && HoldingF.canStrike ) {
                    if (KB.Action)  Actions.attack.Init(Attack.AttackIds.club);
                    if (KB.Action2) Actions.attack.Init(Attack.AttackIds.thrust);
                }
            }
            if (( KB.Left || KB.Right) && !DisabledMoves)
            {
                //  ---
                //  Keys LEFT + RIGHT
                //  ---
                if ( CanWalk )
                {

                    if ((KB.KeyCombos.DoubleRight || KB.KeyCombos.DoubleLeft))
                    {
                        if (FacingLeft == KB.Left)
                        {
                            Actions.dive.Init();
                        }

                        KB.KeyCombos.DoubleRight = KB.KeyCombos.DoubleLeft = false;
                        
                        
                    } else { 
                        if (!IsWalking)
                        {

                            if (FacingLeft != KB.Left) Flip();
                            IsWalking             = true;
                            
                            PBO.OverridePoseIndex = (int)PoseState.Walking;

                        }

                    }
                } 
                else  if (IsCrouching)
                {
                    if (FacingLeft != KB.Left) Flip();
                    //FacingLeft = IsFlipped;
                }
            } 

            else if (KB.Down && !IsCrouching && Actions.prone.state == ProneState.ready && PBO.IsTouchingFloor)
            {
                //  ---
                //  Keys DOWN
                //  ---
                if (KB.KeyCombos.DoubleDown)
                {
                    Actions.prone.Init();
                    KB.KeyCombos.DoubleDown = false;
                } 
                else if (Actions.prone.state == ProneState.ready) { 

                    if (KB.Modifier) PBO.OverridePoseIndex = (int)PoseState.Sitting;
                    else DoCrouch();

                }

            }

            else if (KB.Up)
            {
                //  ---
                //  Keys UP
                //  ---
                if (Actions.prone.state != ProneState.ready)
                {
                    StopPerson();
                    JumpLocked = true;
                } 

                else Actions.jump.Init();
                
            }

            

            else if (KB.Emote)
            {
                DoEmote();
            } else if (IsEmote) {
                IsEmote = false;
                KB.EnableNumberKeys();
                if (EmoteId == -1) Util.Notify("HOLD THIS KEY & CHOOSE EMOTE #", Util.VerboseLevels.Minimal);
            }



            else if (KB.Throw)
            {
                //  ---
                //  Keys Throw
                //  ---
                if (Actions.throwItem.state == ThrowState.ready) Actions.throwItem.Init();

            }


            else {
                
                if (JumpLocked) JumpLocked = false;

            }

            if (KB.Aim)
            {
                //  ---
                //  Keys AIM
                //  ---
                IsAiming = !IsAiming;

                if (IsAiming) AimingInit();

                else AimingStop();

                if (HoldingF != null && !HoldingF.canShoot)
                {
                    Actions.combatMode = !Actions.combatMode;
                    Util.Notify("<color=red>COMBAT MODE: </color>" + (Actions.combatMode ? "ON" : "OFF"), Util.VerboseLevels.Minimal);
                }

            }

            if (KB.ActionHeld || (IsAiming && KB.MouseDown))
            {
                //  ---
                //  Keys ACTION
                //  ---
                if (!IsFiring)
                {
                    // Check if held item is something aimable
                    if (!LB["LowerArmFront"].isActiveAndEnabled || !LB["LowerArmFront"].GripBehaviour.isHolding)
                    {
                        IsFiring = false;
                        return;
                    }

                    IsFiring = true;

                    FireWeapon();
                }
                else if (HoldingF.isAutomatic) FireWeapon();
            }


            
        }


        //
        // ─── DO EMOTE ────────────────────────────────────────────────────────────────
        //
        public void DoEmote()
        {
            if (!IsEmote)
            {
                EmoteId = -1;

                KB.DisableNumberKeys();

                IsEmote = true;
            }

            EmoteId = KB.CheckNumberKey();

            if (EmoteId == -1) return;

            string emotePose;

            switch (EmoteId)
            {
                case 1:
                    emotePose = "JT_Waving";
                    FixBody();
                    break;

                case 2:
                    emotePose = "JT_Victory";
                    break;

                case 3:
                    emotePose = "JT_Prone";
                    break;

                case 4:
                    emotePose = "JT_Dive_1";
                    break;

                case 5:
                    emotePose = "JT_Dive_2";
                    break;

                case 6:
                    emotePose = "JT_Club_1";
                    break;

                case 7:
                    emotePose = "JT_Club_2";
                    break;

                default:
                    emotePose = "JT_Dive_2";
                    break;
            }


            PBO.OverridePoseIndex = PuppetPose[emotePose];
        }

        
        //
        // ─── FIRE WEAPON ────────────────────────────────────────────────────────────────
        //
        private void FireWeapon()
        {
            if (!PBO.IsAlive()) return;

            HoldingF.Activate(HoldingF.isAutomatic);
        }


        //
        // ─── STOP PERSON ────────────────────────────────────────────────────────────────
        //
        public void StopPerson()
        {
            if (!IsWalking && !IsCrouching) { 

                if (PBO.OverridePoseIndex != (int)PoseState.Rest || PBO.OverridePoseIndex != (int)PoseState.Sitting)
                {
                    PBO.OverridePoseIndex = -1;
                }
            }
        }


        //
        // ─── AIMING INIT ────────────────────────────────────────────────────────────────
        //
        public void AimingInit()
        {
            // Check if held item is something that puppet can point
            if (!LB["LowerArmFront"].isActiveAndEnabled || !LB["LowerArmFront"].GripBehaviour.isHolding || HoldingF == null)
            {
                IsAiming = false;
                return;
            }

            IsAiming = HoldingF.canShoot;
            
            if (IsAiming) KB.DisableMouse();


            

            if (HoldingF.holdToSide)
            {
                //HoldingF.SetHoldingPosition();

                LB["LowerArmFront"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                
                HoldingF.ResetPosition(true);
            }


        }

        //
        // ─── AIMING STOP ────────────────────────────────────────────────────────────────
        //
        public void AimingStop()
        {
            IsAiming = false;
            
            LB["LowerArmFront"].Broken = false;
            LB["UpperArmFront"].Broken = false;

            KB.EnableMouse();

            Actions.dive.finale();

            if (HoldingF != null && HoldingF.holdToSide) {

                LB["LowerArmFront"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                HoldingF.ResetPosition();

            }

        }


        //
        // ─── AIM WEAPON ────────────────────────────────────────────────────────────────
        //
        public void AimWeapon()
        {
            //
            //  Reference: https://gamedevbeginner.com/make-an-object-follow-the-mouse-in-unity-in-2d/
            //
            Vector2 mouse = Camera.main.ScreenToWorldPoint(Input.mousePosition);


            if (mouse.x * Facing > (RB2["Head"].position.x * Facing))
            {
                LB["LowerArmFront"].Broken = false;
                LB["UpperArmFront"].Broken = false;

                SetRigidOriginal("LowerArmFront", "inertia");
                SetRigidOriginal("UpperArmFront", "inertia");
                return;

            }

            LB["UpperArmFront"].Broken   = true;
            RB2["UpperArmFront"].inertia = 0.00125f;
            RB2["LowerArmFront"].inertia = 0.00125f;
            
            Vector3 diff = ((Vector3)RB2["LowerArmFront"].position - Camera.main.ScreenToWorldPoint(Input.mousePosition));
            
            //  Prevent spasticated behavior
            float sqrmag = diff.sqrMagnitude;

            if( sqrmag < 105f ) { 
                if (pauseAiming) return;
                if( diff.sqrMagnitude < 101f ) {
                    pauseAiming = true;
                    return;
                }
            }

            pauseAiming = false;
            
            Vector3 angleVelocity;
            if (FacingLeft) angleVelocity = new Vector3(0f, 0f, (Mathf.Atan2(diff.y, diff.x) * Mathf.Rad2Deg) - (85.5f + (16.5f * Facing)));
            else angleVelocity = new Vector3(0f, 0f, (Mathf.Atan2(diff.y, diff.x) * Mathf.Rad2Deg) - (85.5f + (10.5f * Facing)));
            RB2["LowerArmFront"].MoveRotation(Quaternion.Euler(angleVelocity));

        }


        //
        // ─── DO CROUCH ────────────────────────────────────────────────────────────────
        //
        public void DoCrouch()
        {
            if (KB.KeyCombos.DoubleDown)
            {
                Actions.prone.Init();
                IsCrouching = false;
                return;
            }
            else
            {
                PBO.OverridePoseIndex  = PuppetPose["JT_Duck"];
            }
            IsCrouching = true; 
        }


        //
        // ─── INVINCIBLE ────────────────────────────────────────────────────────────────
        //
        public void Invincible(bool doit)
        {
            foreach (LimbBehaviour limb in LB.Values) { limb.ImmuneToDamage = doit; }
        }


        //
        // ─── RUN LIMBS ────────────────────────────────────────────────────────────────
        //
        public void RunLimbs(Action<LimbBehaviour> action)
        {
            foreach (LimbBehaviour limb in LB.Values) { action(limb); }
        }


        //
        // ─── RUN RIGIDS ────────────────────────────────────────────────────────────────
        //
        public void RunRigids(Action<Rigidbody2D> action)
        {
            foreach (Rigidbody2D rigid in RB2.Values) { action(rigid); }
        }
        

        public enum HoldingPositions
        {
            Auto,
            PointingDown,
            PointingForward,
            BothHands,
        }

        //
        // ─── HOLD THING ────────────────────────────────────────────────────────────────
        //
        public void HoldThing(Thing thing)
        {
            ValidateHeldItems();

            if (HoldingF != null)
            {
                DropThing();
                return;
            }

            HoldingF = thing;
            HoldingF.AttachPuppetHand(this, true);

            if (Actions.combatMode && HoldingF.canShoot) Actions.combatMode = false;
            if (IsAiming && !HoldingF.canShoot) IsAiming = false;

        }

        //
        // ─── VALIDATE HELD ITEMS ────────────────────────────────────────────────────────────────
        //
        public void ValidateHeldItems()
        {
            
        }


        public void DropThing()
        {
            //ValidateHeldItems();

            if (HoldingF != null)
            {
                if (HoldingF.P == null)
                {
                    HoldingF = (Thing)null;
                    return;
                }

                if (GripF.isHolding)
                {
                    LB["LowerArmFront"].SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);

                    Util.DisableCollision(PBO.transform, HoldingF.P);

                    //Collider2D[] noCollide = GripF.transform.root.GetComponentsInChildren<Collider2D>();

                    //foreach (Collider2D col1 in HoldingF.tr.root.GetComponentsInChildren<Collider2D>())
                    //{
                    //    foreach (Collider2D col2 in noCollide)
                    //    {
                    //        if ((bool)(UnityEngine.Object)col2 && (bool)(UnityEngine.Object)col1)

                    //        Physics2D.IgnoreCollision(col1, col2);
                    //    }
                    //}

                    PuppetMaster.CheckDisabledCollisions = true;

                    HoldingF.Dropped();
                    HoldingF = (Thing)null;
                }
            }
        }


        //
        // ─── STRENGTHEN ────────────────────────────────────────────────────────────────
        //
        public void Strengthen()
        {
            foreach (KeyValuePair<string, LimbBehaviour> kvp in LB)
            {
                LB[kvp.Key].Vitality             *= 0.1f;
                LB[kvp.Key].RegenerationSpeed    += 5f;
                LB[kvp.Key].ImpactPainMultiplier *= 0.1f;
                LB[kvp.Key].InitialHealth        += 150f;
                LB[kvp.Key].BreakingThreshold    += 10f;
                LB[kvp.Key].ShotDamageMultiplier *= 0.1f;
                LB[kvp.Key].Health               += LB[kvp.Key].InitialHealth;
                LB[kvp.Key].BaseStrength          = Mathf.Min(15f, LB[kvp.Key].BaseStrength + 5f);
            }
        }


        //
        // ─── SET RIGID ORIGINAL ────────────────────────────────────────────────────────────────
        //
        public void SetRigidOriginal(string rigidName, string propName="")
        {
            List<string> props = new List<string>() { "mass", "drag", "inertia" };

            if (propName != "") {
                props.Clear();
                props.Add(propName);
            }

            foreach (string pname in props)
            {
                if (RigidOriginals.TryGetValue(rigidName, out RigidSnapshot rigidOG))
                {
                    switch(pname)
                    {
                        case "mass":
                            RB2[ rigidName ].mass = rigidOG.mass;
                            break;

                        case "drag":
                            RB2[rigidName].drag = rigidOG.drag;
                            break;

                        case "inertia":
                            RB2[rigidName].inertia = rigidOG.inertia;
                            break;
                    }

                }
            }
        }

        public void FixBody()
        {
            RunRigids(BodyInertiaFix);
            foreach (KeyValuePair<string, LimbBehaviour> pair in LB)
            {
                if (LimbOriginals.TryGetValue((string)pair.Key, out LimbSnapshot limbSnapshot))
                {
                    pair.Value.Broken = limbSnapshot.Broken;

                    if (pair.Value.HasJoint)
                    {
                        JointAngleLimits2D jal = pair.Value.Joint.limits;

                        jal.min = limbSnapshot.jLimitMin;
                        jal.max = limbSnapshot.jLimitMax;


                    }

                    pair.Value.Broken = limbSnapshot.Broken;
                }
            }

        }

        public void BodyInertiaFix(Rigidbody2D rigid) => SetRigidOriginal(rigid.name);

        
    }
}
