﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|         
//                                           
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

namespace PuppetMaster
{
    [SkipSerialisation]
    public class PuppetMaster : MonoBehaviour
    {
        [SkipSerialisation] public static bool Activated                      = false;
        [SkipSerialisation] public static bool LastPauseState                 = false;
        [SkipSerialisation] public static bool IsGamePaused                   = false;
        [SkipSerialisation] public static bool CheckDisabledCollisions        = false;

        [SkipSerialisation] public static int ActivePuppet                    = 0;
        
        [SkipSerialisation] public static PhysicalBehaviour LastClickedItem   = null;
        [SkipSerialisation] public static PersonBehaviour LastClickedPerson   = null;
        [SkipSerialisation] public static Puppet Puppet                       = new Puppet();
        [SkipSerialisation] public static ChaseCam ChaseCam                   = null;
        
        private static readonly Dictionary<int, Puppet> PuppetList            = new Dictionary<int, Puppet>();
        

        public static bool CanBePuppet(PersonBehaviour person) => (bool) (person.isActiveAndEnabled && person.IsAlive());


        public void Start()
        {
            KB.SwapBindings(Activated);
        }


        //
        // ─── UNITY UPDATE ────────────────────────────────────────────────────────────────
        //
        public void Update()
        {
            if (InputSystem.Down("PM-TogglePM")) TogglePuppetMode();

            if (!Activated) return;

            if (Global.main.Paused || DialogBox.IsAnyDialogboxOpen || Global.ActiveUiBlock || Global.main.GetPausedMenu() ) 
            {
                if (LastPauseState == false)
                {
                    KB.ReplaceAllKeys();
                    LastPauseState = true;
                    
                    KB.EnableMouse();
                    KB.EnableNumberKeys();
                }
            }
             else if (LastPauseState == true) { 

                LastPauseState = false;
                KB.SwapBindings(true);
                
            }

            KB.CheckKeys();

            if (KB.PClick) CheckNewClicks();

            if (CheckDisabledCollisions && Time.frameCount % 10 == 0) Util.FixDisabledCollisions();

            if ((bool)Puppet?.IsActive) Puppet.CheckControls();

        }

        public void FixedUpdate()
        {
            if ((bool)Puppet?.IsActive) Puppet.FixedUpdate();
        }



        //
        // ─── TOGGLE PUPPET MODE ────────────────────────────────────────────────────────────────
        //
        public static void TogglePuppetMode()
        {
            Activated = !Activated;

            KB.SwapBindings(Activated);

            if (Activated )
            {
                if (ChaseCam == null)
                {
                    //  Init the Chase cam
                    Global.main.gameObject.GetOrAddComponent<ChaseCam>();
                } 
                else
                {
                    //  re-enable the chase cam
                    ChaseCam.enabled = true;
                }

                if (CustomPoses.Poses.Count == 0)
                {
                    // Init the custom poses
                    CustomPoses.InitCustomPoses();
                }

                Util.LoadHoldingPositions();
            } 
            else
            {
                //  Disable the chase cam
                ChaseCam?.Stop();

            }

            Util.Notify("Puppet Master: <color=" + (Activated ? "yellow>On" : "red>Off") + "</color>", Util.VerboseLevels.Minimal);

        }


        //
        // ─── CHECK NEW CLICKS ────────────────────────────────────────────────────────────────
        //
        public static void CheckNewClicks()
        {
            PersonBehaviour LCP = Util.GetClickedPerson();
            if (LCP != null)
            {

                if (LCP == LastClickedPerson)
                {
                    //  This person was clicked twice, so make them the puppet
                    if (CanBePuppet(LCP)) ActivatePuppet(LCP);
                    
                    LCP = null;

                    return;

                }

                LastClickedPerson = LCP;
            }
            else
            {
                PhysicalBehaviour LCI = Util.GetClickedItem();
                if (LCI != null)
                {

                    if (LCI == LastClickedItem)
                    {

                        //  Item was clicked twice, so pick it up automatically
                        if (Puppet != null && Puppet.IsActive ) Puppet.HoldThing(LCI.gameObject.GetOrAddComponent<Thing>());

                        return;
                    }

                    //  Validate clicked item
                    if (CanHoldItem(LCI)) LastClickedItem = LCI;

                }
            }
        }


        //
        // ─── ACTIVATE PUPPET ────────────────────────────────────────────────────────────────
        //
        public static void ActivatePuppet(PersonBehaviour person)
        {
            List<string>ForcedRemoves = new List<string>()
            {
                "ACTIVEHUMAN",
                "EVALTHREAT",
            };
            
            int puppetId = person.GetHashCode();

            //if (ActivePuppet != puppetId) DeactivatePuppet(ActivePuppet);
            ChaseCam.StopChase(puppetId);
            
            if (PuppetList.TryGetValue(puppetId, out Puppet puppet))
            {
                Puppet       = puppet;

                ChaseCam.Chase(
                    Puppet.RB2["Head"].GetComponent<PhysicalBehaviour>(),
                    Puppet.GetHashCode()
                );
            }
            else 
            {
                Puppet = new Puppet();
                Puppet.Init(person);

                PuppetList.Add(puppetId, Puppet);
            }

            //  Disable ActiveHumans for this puppet
            MonoBehaviour[] components = Puppet.PBO.GetComponents<MonoBehaviour>();

            if (components.Length > 0)
            {
                for (int i = components.Length; --i >= 0;)
                {
                    if (ForcedRemoves.Contains(components[i].GetType().ToString().ToUpper()))
                    {
                        Util.Notify("Removed component: " + components[i].GetType().ToString());

                        UnityEngine.Object.DestroyImmediate((UnityEngine.Object)components[i]);
                    }
                }
            }

            Puppet.IsActive        = true;
            ActivePuppet           = puppetId;

        }


        //
        // ─── CAN HOLD ITEM ────────────────────────────────────────────────────────────────
        //
        public static bool CanHoldItem(PhysicalBehaviour item)
        {

            if (!item.Selectable || item.HoldingPositions.Length == 0) return false;

            if (item.TryGetComponent<FreezeBehaviour>(out _)) return false;

            return true;
        }

    }
}
