﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                                    
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;

namespace PuppetMaster
{
    [SkipSerialisation]
    public class Thing : MonoBehaviour
    {
        [SkipSerialisation] public GameObject G;
        [SkipSerialisation] public PhysicalBehaviour P;
        [SkipSerialisation] public Rigidbody2D R;
        [SkipSerialisation] public Transform tr;
        [SkipSerialisation] public Puppet Puppet;
        [SkipSerialisation] public GripBehaviour PuppetGrip;
        [SkipSerialisation] public Rigidbody2D PuppetArm;

        [SkipSerialisation] public string Name;
        [SkipSerialisation] public int Hash;

        [SkipSerialisation] public bool canAim       = false;
        [SkipSerialisation] public bool canShoot     = false;
        [SkipSerialisation] public bool canStab      = false;
        [SkipSerialisation] public bool canStrike    = false;
        [SkipSerialisation] public bool canSlice     = false;
        
        [SkipSerialisation] public bool isAutomatic  = false;
        [SkipSerialisation] public bool holdToSide   = false;
        [SkipSerialisation] public bool hasDetails   = false;
        [SkipSerialisation] public bool isFlashlight = false;
        [SkipSerialisation] public bool isSyringe    = false;
        [SkipSerialisation] public bool twoHands     = false;

        [SkipSerialisation] public float AttackDamage = 0f;
        [SkipSerialisation] public List<int> AttackedList = new List<int>();

        [SkipSerialisation] public float angleHold   = 0f;
        [SkipSerialisation] public float angleAim    = 0f;

        [SkipSerialisation] public float angleOffset = 0f;

        [SkipSerialisation] public float size = 0f;

        [SkipSerialisation] public static Dictionary<int, Vector2> ManualPositions = new Dictionary<int, Vector2>();


        [SkipSerialisation] public float ImpactForceThreshold;

        [SkipSerialisation] public Collider2D Collider;
        [SkipSerialisation] public UnityEvent Actions;

        
        [SkipSerialisation] public Vector2 HoldingPosition;
        [SkipSerialisation] public Vector3 HoldingOffset;

        [SkipSerialisation] public bool IsFlipped => (bool)(tr.localScale.x < 0.0f);

        private string sortingLayerName;
        private int sortingOrder;


        public void Awake()
        {
            G  = gameObject;
            P  = gameObject.GetComponent<PhysicalBehaviour>();
            R  = gameObject.GetComponent<Rigidbody2D>();
            tr = P.transform;

            Name = P.name;
            Hash = P.GetHashCode();


            sortingLayerName = P.GetComponent<SpriteRenderer>().sortingLayerName;
            sortingOrder     = P.GetComponent<SpriteRenderer>().sortingOrder;

            GetDetails();
        }

        //
        // ─── COLLISIONS ────────────────────────────────────────────────────────────────
        //
        private void OnCollisionEnter2D(Collision2D coll=null)
        {
            //  if AttackDamage is not set, then we're not attacking
            if (AttackDamage <= 0f) return;

            if (coll == null) return;
            LimbBehaviour theHitLimb = (LimbBehaviour)null;
            PersonBehaviour hitperson = coll.rigidbody.GetComponentInParent<PersonBehaviour>();
            
            if (hitperson != null)
            {

                if (AttackedList.Contains(hitperson.GetHashCode())) return;
                if (canStab)
                {
                    if (coll.rigidbody.TryGetComponent<LimbBehaviour>(out LimbBehaviour limbStabbed))
                    {
                        theHitLimb = limbStabbed;
                        //  Check if power was enough to do a finisher
                        if (Puppet.Actions.attack.CurrentMove == Attack.MoveTypes.slash)
                        {
                            if (AttackDamage > 30 && AttackDamage < 40)
                            {
                                int health = Util.GetOverallHealth(hitperson);
                                if (health > 75) health = 75;
                                if (UnityEngine.Random.Range(1, 100) > health)
                                {
                                    //Util.Detach(hitperson, limbStabbed);
                                    limbStabbed.Slice();
                                    Util.Notify("Critical", Util.VerboseLevels.Minimal);
                                }
                            }
                        }

                        else if (Puppet.Actions.attack.CurrentMove == Attack.MoveTypes.stab)
                        { 
                            if (AttackDamage > 30 && AttackDamage < 40)
                            {
                                int health = Util.GetOverallHealth(hitperson);
                                if (health > 75) health = 75;
                                if (UnityEngine.Random.Range(1, 100) > health)
                                {
                                    if (coll.rigidbody.TryGetComponent<PhysicalBehaviour>(out PhysicalBehaviour PB))
                                    {
                                        Stabbing stabbing = new Stabbing(P, PB, (Vector2)(tr.position - PB.transform.position).normalized, coll.collider.ClosestPoint(tr.position));

                                        limbStabbed.Stabbed(stabbing);
                                        Util.Notify("Critical", Util.VerboseLevels.Minimal);
                                    }
                                }
                            }

                        }
                    }
                }
                else
                {
                    if (coll.rigidbody.TryGetComponent<LimbBehaviour>(out LimbBehaviour limbStabbed))
                    {
                        theHitLimb = limbStabbed;
                        //  Check if power was enough to do a finisher
                        if (Puppet.Actions.attack.CurrentMove == Attack.MoveTypes.slash)
                        {
                            if (AttackDamage > 30 && AttackDamage < 40)
                            {
                                int health = Util.GetOverallHealth(hitperson);
                                if (health > 75) health = 75;
                                if (UnityEngine.Random.Range(1, 100) > health)
                                {
                                    //Util.Detach(hitperson, limbStabbed);
                                    limbStabbed.Crush();
                                    Util.Notify("Critical", Util.VerboseLevels.Minimal);
                                }
                            }
                        }

                        else if (Puppet.Actions.attack.CurrentMove == Attack.MoveTypes.stab)
                        {
                            if (AttackDamage > 30 && AttackDamage < 40)
                            {
                                int health = Util.GetOverallHealth(hitperson);
                                if (health > 75) health = 75;
                                if (UnityEngine.Random.Range(1, 100) > health)
                                {
                                    if (coll.rigidbody.TryGetComponent<PhysicalBehaviour>(out PhysicalBehaviour PB))
                                    {
                                        limbStabbed.BreakBone();
                                        Util.Notify("Critical", Util.VerboseLevels.Minimal);
                                    }
                                }
                            }

                        }

                    }
                }

                AttackedList.Add(hitperson.GetHashCode());

                hitperson.PainLevel  = 0f;
                hitperson.ShockLevel = 0f;
                if (theHitLimb != null)
                {
                    if (theHitLimb.Health  < theHitLimb.InitialHealth)  theHitLimb.Health += (theHitLimb.InitialHealth - theHitLimb.Health) / 2f;
                    if (theHitLimb.BruiseCount > 2) theHitLimb.BruiseCount--;
                    
                }

                //coll.rigidbody.AddForce(R.velocity * Puppet.TotalWeight * AttackDamage * UnityEngine.Random.Range(0.3f, 0.5f));
            }

            //this.Actions.Invoke();
        }

        //
        // ─── HELD BY ────────────────────────────────────────────────────────────────
        //
        public PersonBehaviour HeldBy()
        {
            if (!P.beingHeldByGripper) return (PersonBehaviour) null;

            GripBehaviour[] Grips = Global.FindObjectsOfType<GripBehaviour>();

            foreach (GripBehaviour grip in Grips)
            {
                if (grip != null && grip.CurrentlyHolding == P)
                {
                    return grip.GetComponentInParent<PersonBehaviour>();
                }
            }

            return (PersonBehaviour) null;
        }


        //
        // ─── BREAK CONNECTIONS ────────────────────────────────────────────────────────────────
        //
        public void BreakConnections(bool clearPuppet=false)
        {

            GripBehaviour[] Grips = Global.main.gameObject.GetComponentsInChildren<GripBehaviour>();

            foreach (GripBehaviour grip in Grips)
            {
                if (grip.CurrentlyHolding != P) continue;

                // Break joint
                FixedJoint2D ItemJoint;
                        
                while (ItemJoint = grip.gameObject.GetComponent<FixedJoint2D>())
                {
                    UnityEngine.Object.DestroyImmediate(ItemJoint);
                }

            }

            P.beingHeldByGripper = false;


            if (clearPuppet)
            {
                if (Puppet.GripF.CurrentlyHolding == P) { Puppet.GripF.CurrentlyHolding = null; Puppet.GripF.isHolding = false; }
                if (Puppet.GripB.CurrentlyHolding == P) { Puppet.GripB.CurrentlyHolding = null; Puppet.GripF.isHolding = false; }
                    
                Puppet.HoldingF = (Thing)null;
                Puppet.IsAiming = false;
                Puppet          = null;
                PuppetGrip      = null;
            }
            
        }


        //
        // ─── DETACH ────────────────────────────────────────────────────────────────
        //
        public void Dropped()
        {
            if (Puppet != null)
            {
                //if (Puppet.HoldingF != null)
                //{
                //    //Puppet.IsAiming = false;
                //    //P.MakeWeightful();
                //    //Util.DisableCollision(Puppet.PBO.transform, P);
                //}

                Puppet     = null;
                PuppetGrip = null;

                LayerFix(false);
            }
            P.MakeWeightful();
        }

        //
        // ─── ATTACH PUPPET HAND ────────────────────────────────────────────────────────────────
        //
        public bool AttachPuppetHand(Puppet puppet, bool frontHand = true)
        {
            if (!hasDetails) GetDetails();

            P.MakeWeightless();
            
            Puppet                = puppet;
            PuppetArm             = frontHand ? Puppet.RB2["LowerArmFront"] : Puppet.RB2["LowerArm"];
            PuppetGrip            = frontHand ? Puppet.GripF : Puppet.GripB;
            GripBehaviour AltGrip = frontHand ? Puppet.GripB : Puppet.GripF;
            HoldingPosition       = new Vector3( float.MaxValue, 0.0f );

            foreach (Vector3 HPos in P.HoldingPositions) if (HPos.x < HoldingPosition.x) HoldingPosition = HPos;
            SetHoldingPosition(HoldingPosition);

            if (AltGrip.isHolding)
            {
                PhysicalBehaviour AltItem = AltGrip.CurrentlyHolding;

                foreach (Collider2D col1 in AltItem.transform.root.GetComponentsInChildren<Collider2D>())
                {
                    foreach (Collider2D col2 in tr.root.GetComponentsInChildren<Collider2D>())
                    {
                        if ((bool)(UnityEngine.Object)col2 && (bool)(UnityEngine.Object)col1)
                            Physics2D.IgnoreCollision(col1, col2);
                    }
                }
            }

            ResetPosition();

            //  Adjust layers of the hand and the held object so object is over body but under hand and arms
            LayerFix(true);

            if (frontHand) Puppet.HoldingF = this;

            return true;

        }

        public void LayerFix(bool inHand)
        {
            if (inHand)
            {
                int SOrder = Puppet.RB2["LowerArmFront"].GetComponent<SpriteRenderer>().sortingOrder;
                if (SOrder <= 1)
                {
                    SOrder = 2;
                    Puppet.RB2["LowerArmFront"].GetComponent<SpriteRenderer>().sortingOrder = SOrder;
                    Puppet.RB2["UpperArmFront"].GetComponent<SpriteRenderer>().sortingOrder = SOrder;
                }

                P.GetComponent<SpriteRenderer>().sortingLayerName = Puppet.RB2["LowerArmFront"].GetComponent<SpriteRenderer>().sortingLayerName;
                P.GetComponent<SpriteRenderer>().sortingOrder = SOrder + -1;
            }
            else
            {
                P.GetComponent<SpriteRenderer>().sortingLayerName = sortingLayerName;
                P.GetComponent<SpriteRenderer>().sortingOrder     = sortingOrder;
            }
        }


        //
        // ─── RESET POSITION ────────────────────────────────────────────────────────────────
        //
        public void ResetPosition(bool norotate=false)
        {

            
            if (Puppet.IsFlipped != IsFlipped) Flip();

            if (angleHold == 0f)
            {
                angleHold = (holdToSide && !norotate) ? 5.0f : 95.0f;
                angleAim  = 95.0f;
            }

            //if (P.Properties.SharpAxes.Length == 1) angleHold = angleAim = P.Properties.SharpAxes[0].Axis.x;

            float thingRotation = Puppet.IsAiming ? angleAim : angleHold;

            thingRotation += angleOffset;

            tr.rotation = Quaternion.Euler(0.0f, 0.0f, IsFlipped ? PuppetArm.rotation + thingRotation : PuppetArm.rotation - thingRotation);

            tr.position += PuppetGrip.transform.TransformPoint(PuppetGrip.GripPosition) - tr.TransformPoint((Vector3)HoldingPosition);

            PuppetArm.SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);

            if (angleAim != angleHold)
            {
                tr.rotation = Quaternion.Euler(0.0f, 0.0f, IsFlipped ? PuppetArm.rotation + thingRotation : PuppetArm.rotation - thingRotation);
            }

            //if (HoldingOffset != null)
            //{
            //    ModAPI.Notify("offset");
            //    tr.position -= PuppetGrip.transform.TransformPoint(PuppetGrip.GripPosition) - tr.TransformPoint((Vector3)HoldingOffset);
            //}

        }


        

        //
        // ─── FLIP ────────────────────────────────────────────────────────────────
        //
        public void Flip()
        {
            Vector3 scale = tr.localScale;
            scale.x      *= -1;
            tr.localScale = scale;
        }
        

        //
        // ─── ACTIVATE ────────────────────────────────────────────────────────────────
        //
        public void Activate(bool continuous=false)
        {
            P.SendMessage(continuous ? "UseContinuous" : "Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
        }


        //
        // ─── SET HOLDING POSITION ────────────────────────────────────────────────────────────────
        //
        public void SetHoldingPosition(Vector3 newPosition, bool save=false)
        {
            //  We call this when its detected player has repositioned an item
            int hoPo = P.name.GetHashCode();

            if (!save && ManualPositions.ContainsKey(hoPo))
            {
                HoldingOffset = ManualPositions[hoPo];
                return;
            }

            HoldingOffset = newPosition;


            if (save) {
                

                if (ManualPositions.ContainsKey(hoPo)) ManualPositions[hoPo] = HoldingOffset;
                else ManualPositions.Add(hoPo, HoldingOffset);

                Util.SaveHoldingPositions();

            }
            
        }

        public void GetPositionChange()
        {
            //  fuckthisfuckingfuckitybullshitnonworkingconfusingassholecodebolloxeddicklickingtubbybitchdontusedrugsreadafuckingbook
            //float fdirec = IsFlipped ? 1.0f : -1.0f;

            //Vector3 t2 = new Vector3(HoldingPosition.x, HoldingPosition.y);

            //Vector3 nhp = new Vector3(float.MaxValue, 0.0f);
            //foreach (Vector3 HPos in P.HoldingPositions) if (HPos.x < nhp.x) nhp = HPos;

            //float rotation = PuppetArm.rotation + (95.0f * fdirec);
            
            //ModAPI.Notify("hpx: " + nhp.x + " hpY: " + nhp.y);
            //ModAPI.Notify("Rotation: " + rotation);
            
            //Vector3 mypos = PuppetArm.transform.InverseTransformPoint(PuppetArm.transform.position) - tr.InverseTransformPoint(R.transform.position);
            //ModAPI.Notify("X1: " + mypos.x + " Y1: " + mypos.y);

            ////tr.position += PuppetArm.transform.TransformPoint(PuppetArm.position) - tr.TransformPoint((Vector3)mypos);
            //tr.position += PuppetArm.transform.TransformPoint(PuppetArm.position) - tr.TransformPoint((Vector3)mypos);

            //// x: -0.4675879   y: -0.04876827
            //Vector3 offsetp = new Vector3(-0.4752571f, -1.795648f);


            //tr.position = PuppetGrip.transform.position - offsetp;


            //tr.position = PuppetGrip.transform.position + PuppetGrip.transform.TransformPoint(PuppetGrip.GripPosition) - tr.TransformPoint((Vector3)offsetp);

            //tr.rotation = Quaternion.Euler(0.0f, 0.0f, IsFlipped ? PuppetArm.rotation + 95.0f : PuppetArm.rotation - 95.0f);


            //Vector3 t3 = new Vector3(tr.TransformPoint(HoldingPosition).x, tr.TransformPoint(HoldingPosition).y);
            //Vector3 t3 = tr.TransformPoint((Vector3)HoldingPosition);
            //ModAPI.Notify("X1: " + t3.x + " Y1: " + t3.y);

            //Vector3 t5 = PuppetGrip.GripPosition - tr.position;
            //ModAPI.Notify("X2: " + t5.x + " Y2: " + t5.y);

            //Vector3 t4 = tr.InverseTransformPoint(t5);
            //ModAPI.Notify("X2: " + t4.x + " Y2: " + t4.y);



            //tr.rotation = Quaternion.Euler(0.0f, 0.0f, IsFlipped ? PuppetArm.rotation + thingRotation : PuppetArm.rotation - thingRotation);

            //tr.position += PuppetGrip.transform.TransformPoint(PuppetGrip.GripPosition) - tr.TransformPoint((Vector3)HoldingPosition);


            //HoldingPosition = new Vector3(float.MaxValue, 0.0f);
            //foreach (Vector3 HPos in P.HoldingPositions) if (HPos.x < HoldingPosition.x) HoldingPosition = HPos;
            //SetHoldingPosition(HoldingPosition);


        }


        //
        // ─── GET DETAILS ────────────────────────────────────────────────────────────────
        //
        public void GetDetails()
        {
            const float LargeItemLength = 1.1f;

            hasDetails = true;

            //  Define Guns and Rate of fire mode
            //
            Type[] ShootersAuto = new Type[] {
                typeof(FlamethrowerBehaviour),
                typeof(PhysicsGunBehaviour),
                typeof(MinigunBehaviour),
                typeof(TemperatureRayGunBehaviour),
                // typeof(FireExtinguisherBehaviour),  ?obsolete
            };

            Type[] ShootersManual = new Type[]
            {
                typeof(BeamformerBehaviour),
                typeof(RocketLauncherBehaviour),
                typeof(GenericScifiWeapon40Behaviour),
                typeof(LightningGunBehaviour),
                typeof(PulseDrumBehaviour),
                typeof(HealthGunBehaviour),
                typeof(FreezeGunBehaviour),
                typeof(SingleFloodlightBehaviour),
            };

            if (P.name.Contains("Syringe")) {
                
                angleOffset = -95f;
                isSyringe   = true;   

            }


            foreach (Type type in ShootersAuto)
            {
                if (P.TryGetComponent(type, out _))
                {
                    canShoot    = true;
                    isAutomatic = true;
                    break;
                }
            }

            foreach (Type type in ShootersManual)
            {
                if (P.TryGetComponent(type, out _))
                {
                    canShoot    = true;
                    isAutomatic = false;
                    break;
                }
            }

            //  @Todo: need to return modified values to their OG after they're dropped
            
            //  Determine if we can do auto firing

            if (P.TryGetComponent<FirearmBehaviour>(out FirearmBehaviour FBH))
            {
                canShoot             = true;
                isAutomatic          = FBH.Automatic;
                FBH.Cartridge.Recoil = 0.1f;
            }
            else if (P.TryGetComponent<ProjectileLauncherBehaviour>(out ProjectileLauncherBehaviour PLB))
            {
                canShoot             = true;
                isAutomatic          = PLB.IsAutomatic;
                PLB.recoilMultiplier = 0.1f;
            }
            else if (P.TryGetComponent<BlasterBehaviour>(out BlasterBehaviour BB))
            {
                canShoot             = true;
                isAutomatic          = BB.Automatic;
                BB.Recoil            = 0.1f;
            }
            else if (P.TryGetComponent<BeamformerBehaviour>(out BeamformerBehaviour BB2))
            {
                BB2.RecoilForce      = 0.1f;
            }
            else if (P.TryGetComponent<GenericScifiWeapon40Behaviour>(out GenericScifiWeapon40Behaviour GWB))
            {
                GWB.RecoilForce = 0.1f;
            }
            //else if (P.TryGetComponent<AcceleratorGunBehaviour>(out AcceleratorGunBehaviour AB))
            //{
            //    AB.RecoilIntensity = 0.1f;
            //    canShoot           = true;
            //    isAutomatic        = true;
            //}
            else if (P.TryGetComponent<ArchelixCasterBehaviour>(out _))
            {
                canShoot    = true;
                isAutomatic = false;
            }
            else if (P.TryGetComponent<SingleFloodlightBehaviour>(out _))
            {
                isFlashlight = true;
                angleHold    = 95f;
                angleAim     = 180f;
            }


            

            //  Flag Larger Items since they interfere with walking
            holdToSide =  canShoot && P.ObjectArea > LargeItemLength;
            size       = Mathf.Max(tr.lossyScale.x, tr.lossyScale.y);

            //  Check for clubs/bats
            if (!canShoot)
            {
                canStab     = P.Properties.Sharp && P.StabCausesWound;
                canStrike   = true;
                twoHands    = size > 1.1f;
                canSlice    = P.TryGetComponent<SharpOnAllSidesBehaviour>(out _);
            }

        }


        public void Destroy()
        {
            Puppet?.DropThing();
            
            BreakConnections();

            if (P.Deletable)
            {
                G.SendMessage("OnUserDelete", SendMessageOptions.DontRequireReceiver);

                UnityEngine.Object.Destroy((UnityEngine.Object)G);
            }

            P = (PhysicalBehaviour)null;
            R = (Rigidbody2D)null;
            G = (GameObject)null;
        }
        
    }
}
